/*
 * Help Access Library 
 * A CGI-Program to read Windows Help files in a Web Browser.
 *
 * Copyright (C) 1995-2000 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* dib2png.c - Part of WinHelpCGI
  
   Converts a Graphic in Windows DIB format into a PNG file.
  
   Interesting limitations:
   - SHG-Files built from Metafiles cannot be processed as we'd have to add
     code to convert the Metafile to a Bitmap.
   - MAPs/SHGs are not supported.

   (C)1996-2000 Herd Software Development, Kettelerstr. 35 / 68642 Buerstadt / Germany"
*/


#include <windows.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <png.h>                // libpng2-dev
#include "stringbuffer.h"       // STRINGBUFFER Appendable strings structures
#include "dib2png.h"            // Our own private header file.


static void user_write_data(png_structp png_ptr, png_bytep data, png_size_t length);


/* Dib2Png
 *
 * A simple PNG-Creation procedure using libpng.
 */
void Dib2Png(STRINGBUFFER *sbout, HGLOBAL hDIB)
{
  png_structp  png_ptr;        		// Give Access to PNG data.
  png_infop    info_ptr;
  png_bytep    row_pointer;

  LPBITMAPINFOHEADER lpbmi;       	/* Pointer to the Windows DIB structure */

  int          color_type;
  unsigned int palette_count=0;

  png_colorp   palette = NULL;

  unsigned int src_bytes_per_line;
  unsigned int image_line;

  /* Create and initialize the png_struct. We also check that
	* the library version is compatible with the one used at compile time,
	* in case we are using dynamically linked libraries.  REQUIRED.
	*/
  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (png_ptr == NULL)
	 {
      return;
	 }

  /* Allocate/initialize the image information data.  REQUIRED */
  info_ptr = png_create_info_struct(png_ptr);
  if (info_ptr == NULL)
	 {
      png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
      return;
	 }

  /* Set error handling.  REQUIRED if you aren't supplying your own
	* error hadnling functions in the png_create_write_struct() call.
	*/
  if (setjmp(png_ptr->jmpbuf))
	 {
      /* If we get here, we had a problem writing the file */
      png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
      return;
	 }
  
  /* Set the I/O callback function to transfer directly to memory */
  png_set_write_fn(png_ptr, (voidp) sbout, user_write_data, NULL);
  
  /* Get access to the windows DIB data */
  lpbmi = (LPBITMAPINFOHEADER) GlobalLock(hDIB);
  assert( lpbmi != NULL );

  /* Find out about the color type */
  color_type = lpbmi->biBitCount > 8 
	 ? PNG_COLOR_TYPE_RGB
	 : PNG_COLOR_TYPE_PALETTE;
	 
  /* Set the image information here.  Width and height are up to 2^31,
	* bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
	* the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
	* PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
	* or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
	* PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
	* currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
	*/
  png_set_IHDR(
					png_ptr, 
					info_ptr, 
					lpbmi->biWidth, 
					lpbmi->biHeight, 
					color_type == PNG_COLOR_TYPE_RGB ? 8 : lpbmi->biBitCount,
					color_type,
					PNG_INTERLACE_NONE,
					PNG_COMPRESSION_TYPE_BASE,
					PNG_FILTER_TYPE_BASE
					);

  /* Take care about the color palette */
  if (color_type == PNG_COLOR_TYPE_PALETTE)
	 {
		unsigned int palette_index;
		LPRGBQUAD lprgbq = (LPRGBQUAD) ( ((char *) lpbmi) + lpbmi->biSize);

		palette_count = lpbmi->biClrUsed ? lpbmi->biClrUsed : (1<<lpbmi->biBitCount);

		/* set the palette if there is one.  REQUIRED for indexed-color images */
		palette = (png_colorp)png_malloc(png_ptr, palette_count * sizeof (png_color));

		assert(palette != NULL);

		/* Copy the color palette */
		for (palette_index=0; 
			  palette_index<palette_count;
			  palette_index++)
		  {
			 palette[palette_index].blue =lprgbq->rgbBlue;
			 palette[palette_index].green=lprgbq->rgbGreen;
			 palette[palette_index].red  =lprgbq->rgbRed;
			 lprgbq++;
		  }

		/* ... set palette colors ... */
		png_set_PLTE(png_ptr, info_ptr, palette, palette_count);
	 } // color_type = PNG_COLOR_TYPE_PALETTE

  if (lpbmi->biXPelsPerMeter!=0)
  {
	 png_set_pHYs(
					  png_ptr,
					  info_ptr,   
					  lpbmi->biXPelsPerMeter,
					  lpbmi->biYPelsPerMeter,
					  PNG_RESOLUTION_METER
					  );
  }

  /* Write the file header information. */
  png_write_info(png_ptr, info_ptr);

  if (lpbmi->biBitCount==32)
  {
	 png_set_filler(png_ptr, 0, PNG_FILLER_AFTER);   /* Use 4 Source bytes to create one pixel */
  }

  if (lpbmi->biBitCount>=24)
  {
	 png_set_bgr(png_ptr);
  }

  /* Compute DIB metrics */
  src_bytes_per_line = (lpbmi->biWidth * lpbmi->biBitCount + 31) / 32 * 4;

  /* Loop over scan lines. */
  for (image_line=0;
		 image_line<lpbmi->biHeight;
		 image_line++)
	 {
		row_pointer = 
		  ((unsigned char *) lpbmi)
		  + lpbmi->biSize
		  + palette_count * sizeof(RGBQUAD)
		  + (lpbmi->biHeight-image_line-1) * src_bytes_per_line;

		png_write_row(png_ptr, row_pointer);
	 }

  /* Write the end of the PNG file */
  png_write_end(png_ptr, info_ptr);
  
  /* Clean up */
  png_destroy_write_struct(&png_ptr, &info_ptr);
}


/**
 * User-Supplied callback function to write Data
 */
static void user_write_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
  STRINGBUFFER * sbout = (STRINGBUFFER *) png_get_io_ptr(png_ptr);
  sbwrite(sbout, data, length);
}
