/*
 * Help Access Library 
 * A CGI-Program to read Windows Help files in a Web Browser.
 *
 * Copyright (C) 1995-2004 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* hlp2tar.c -- part of WinHelpCGI

   This file contains code to create a unix "TAR" archive from
   the Helpfiles given on the command line.

   (C)1996-2004 Herd Software Development, Kettelerstr. 35 / 68642 Buerstadt / Germany"
*/


#include <windows.h>
#include <fcntl.h>
#include <dirent.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include "hlpacces.h"           // Help Access Library Functions
#include "winhelpcgi.h"         // Header-File to connect the pieces of WinHelpCGI application
#include "virtchmfs.h"          // Virtual CHM filesystem


#ifdef ENABLE_NLS
#include <libintl.h>

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)
#else
#define _(String) (String)
#endif

/// Data extracted from Command line parameters
struct cmdparams
{
  /// Set to TRUe on Parameter --htmlonly to avoid generating .png, and .hh=? files.
  BOOL htmlonly;

  /// Add the title of the help file to the titles of the topic files.
  BOOL addhlptitle;
};


/// Data used during enumeration

typedef struct
{
  /// Virtual CHM-Filesystem opened by VCHMOpen()
  struct vchm *pchm;

  /// Filename of the .hlp file
  const char *HlpFileName;

  /// Parameters extracted from the command line
  struct cmdparams *params;
} ENUMDATA;



/**
 * Copy tfile data into a tar header format.
 *
 * @param sb          STRINGBUFFER structure containing the Data to be written
 * @param name        Name that the filw should have in the tar.
 */
static
void
_WriteToTar(STRINGBUFFER *sb, const char *name)
{
  int size = sblength(sb);
  int remainder = 512 - (size % 512);
  char tarheader[512];

  memset(tarheader, 0, sizeof(tarheader));
  snprintf(tarheader+0, 100, name);
  strcpy(tarheader+100, "0100644"); // file mode
  strcpy(tarheader+100+8, "0000000"); // File user ID
  strcpy(tarheader+100+8+8, "0000000"); // File group ID
  sprintf(tarheader+100+8+8+8, "%011o", size); // FIle size
  sprintf(tarheader+100+8+8+8+12, "%011o", (int) time(NULL)); // Modification time
  memset(tarheader+100+8+8+8+12+12, ' ', 8); // init checksum
  strcpy(tarheader+100+8+8+8+12+12+8, "0"); // Link flag

  {
    int i;
    int sum;
    for (i=sum=0; i<512; i++)
      {
        sum += tarheader[i];
      }

    sprintf(tarheader+100+8+8+8+12+12, "%07o", sum); // checksum
  }

  fwrite(tarheader, 1, 512, stdout);

  sbflush(sb, stdout);

  if (remainder!=512)
    {
      memset(&tarheader, 0, remainder);
      fwrite(&tarheader, 1, remainder, stdout);
    }
}



/**
 * Send a virtual file as tar contents to stdout
 *
 * @param pchm          Virtual CHM Filesystem handle
 * @param HlpFileName   Filename of the .hlp file.
 * @param virtfilename  Filename of the virtual file.
 * @param formatflags   Additional flags for formatting of the HTML files.
 *
 * @return TRUE on success
 */
static
BOOL
_VirtFileToTar(struct vchm *pchm, const char *HlpFileName, const char *virtfilename, unsigned int formatflags)
{
  BOOL success = FALSE;
  char tarfilename[500];
  STRINGBUFFER sbout;
  char mime_type[50];

  snprintf(tarfilename, sizeof(tarfilename),
           "%s/%s",
           HlpFileName, virtfilename);

  memset(&sbout, 0, sizeof(sbout));
  success =
    VCHMGetFile(pchm, &sbout, virtfilename, formatflags, mime_type, sizeof(mime_type));

  if (success)
    {
      _WriteToTar(&sbout, tarfilename);
    }
  
  return success;
}


/**
 * Convert the images found in the .HLP file to .tar format on stdout
 *
 * @param pchm          Virtual CHM Filesystem handle
 * @param HlpFileName   Filename of the windows help file.
 */
static
void
_ImagesToTar(struct vchm *pchm, const char *HlpFileName)
{
  BOOL cont;
  int i;
  
  for (i=0, cont=TRUE; cont; i++)
    {
      char virtfilename[200];
      
      snprintf(virtfilename, sizeof(virtfilename),
               "png/bm%d.png",
               i);
      
      cont = _VirtFileToTar(pchm, HlpFileName, virtfilename, 0);
    }
}


/**
 * Include the HTML-files listed in the [files] section
 * of the .hhp file into the .tar file.
 *
 * @param dta       Temporary Data fields used.
 */
static
void
_ListedFilesToTar(ENUMDATA *dta)
{
  char mimetype[50];
  STRINGBUFFER sbhhp;
  memset(&sbhhp, 0, sizeof(sbhhp));
  if (VCHMGetFile(dta->pchm, &sbhhp, "*.hhp", 0, mimetype, sizeof(mimetype)))
    {
      char line[500];
      char *dst;
      const char *src=sbget(&sbhhp);
      BOOL in_files_section = FALSE;
      while (*src)
        {
          // Scan one line into the line buffer
          while (*src && isspace(*src)) src++;

          // 
          for (dst=line; (dst-line)<sizeof(line)-1 && *src && *src!='\r' && *src!='\n'; )
            {
              *dst++ = *src++;
            }
          *dst=0;

          if (line[0]=='[')
            {
              in_files_section = !strcasecmp(line, "[files]");
            }
          else if (in_files_section && line[0])
            {
              /* We've got a file name in the line buffer */
              _VirtFileToTar(dta->pchm, dta->HlpFileName, line, dta->params->addhlptitle ? HTMLTOPIC_ADDHLPTITLE : 0);
            }
        }
    }
  else
    {
      fprintf(stderr, _("Can't open the virtual .hhp file. File \"%s\" might not be a windows .hlp file\n"), dta->HlpFileName);
    }
} 




/**
 * Convert one help file given on the command line to
 * a set of HTML files in tar format on stdout.
 *
 * @param HlpFileName Filename of the help file to be converted
 * @param params      Parameters extracted from the command line arguments
 *
 * @return EXIT_SUCCESS
 */
static
int
_Hlp2Tar(const char *HlpFileName, struct cmdparams *params)
{
  int result = EXIT_FAILURE;
  ENUMDATA dta;
  dta.HlpFileName = HlpFileName;
  dta.params = params;
  
  if (NULL!=(dta.pchm = VCHMOpen(dta.HlpFileName)))
    {
      _ListedFilesToTar(&dta);

      if (!params->htmlonly)
        {
          _ImagesToTar(dta.pchm, HlpFileName);

          _VirtFileToTar(dta.pchm, HlpFileName, "winhelpcgi.hhp", 0);
          _VirtFileToTar(dta.pchm, HlpFileName, "winhelpcgi.hhc", 0);
          _VirtFileToTar(dta.pchm, HlpFileName, "winhelpcgi.hhk", 0);
          _VirtFileToTar(dta.pchm, HlpFileName, "winhelpcgi.h", 0);
        }

      VCHMClose(dta.pchm);
    }
  else
    {
      fprintf(stderr, _("Cannot open Help file %s reason: %s\n"), HlpFileName, strerror(errno) );
    }
  return result;
}
  


/**
 * Recursively collect help files.
 * 
 * @param dirname     Directory name to continue recursion
 * @param params      Parameters extracted from the command line arguments
 */
static
void
_ScanDirectoryForHelpFiles(const char *dirname, struct cmdparams *params)
{
  DIR           * dir=opendir(dirname);
  if (dir != NULL)
    {
      char fullname[100];
      struct dirent * ent;
      struct stat st;
      while (NULL!=(ent=readdir(dir)))
        {
          snprintf(fullname, sizeof(fullname), "%s/%s", dirname, ent->d_name);
          if (stat(fullname, &st)==0)
            {
              HIFS           ifs;
              
              if (S_ISREG(st.st_mode) &&
                  NULL!=(ifs = IFSOpen(fullname)))
                {
                  IFSClose(ifs);
                  _Hlp2Tar(fullname, params);
                } 
              else if (S_ISDIR(st.st_mode) && ent->d_name[0]!='.')
                {
                  /* Recurse... */
                  _ScanDirectoryForHelpFiles(fullname, params);
                }
            }
          
        }
      closedir(dir);
    }
  else
    {
      fprintf(stderr, _("Cannot open directory \"%s\" reason: %s\n"), dirname, strerror(errno));
    }
}



/**
 * Convert a set of help files given on the command line to
 * a set of HTML files in tar format on stdout.
 *
 * @param argc   Number of Help files from command line
 * @param argv   List of Help file names
 *
 * @return EXIT_SUCCESS
 */
int
Hlp2Tar(int argc, char **argv)
{
  struct cmdparams params;
  memset(&params, 0, sizeof(params));

  while (argc>0 && argv[0][0]=='-')
    {
      if (!strcasecmp(argv[0], "--htmlonly"))
        {
          params.htmlonly = TRUE;
        }
      else if (!strcasecmp(argv[0], "--addhlptitle"))
        {
          params.addhlptitle = TRUE;
        }
      argc--;
      argv++;
    }

  if (argc==0)
    {
      _ScanDirectoryForHelpFiles("helpfiles", &params);
    }
  else
    {
      int i;
      for (i=0; i<argc; i++)
        {
          struct stat st;
          if (stat(argv[i], &st)==0)
            {
              if (S_ISDIR(st.st_mode))
                {
                  _ScanDirectoryForHelpFiles(argv[i], &params);
                }
              else
                {
                  _Hlp2Tar(argv[i], &params);
                }
            }
        }
    }

  return EXIT_SUCCESS;
}
