/*
 * Help Access Library 
 * A CGI-Program to read Windows Help files in a Web Browser.
 *
 * Copyright (C) 1995-2004 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* includes. -- part of WinHelpCGI

   This piece contains functions to handle inclusion of formatting data
   stored in files outside of the .hlp files. 

   This is useful for online publishing purposes.
 
   (C)1996-2004 Herd Software Development, Kettelerstr. 35 / 68642 Buerstadt / Germany"
*/

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <sys/wait.h>
#include <unistd.h>
#ifdef HAVE_SIGNAL_H
#include <signal.h> // LINUX, MacOS X
#elif defined(HAVE_SYS_SIGNAL_H)
#include <sys/signal.h> // AIX
#endif
#include "hlpacces.h"
#include "winhelpcgi.h"
#include <errno.h>


/**
 * Get a suitable language code "de", "nl", "fr" or "en" from the HTTP_ACCEPT_LANGUAGE
 * environment variable.
 *
 * @param  pLanguage         Pointer to a buffer that will hold the language code on return
 * @param  buffer_size       Size of the buffer in pLanguage
 *
 * @return TRUE if suitable language has been found
 */
BOOL
ChooseLanguage(LPSTR pLanguage, int buffer_size)
{
  const char *accept_language = getenv("HTTP_ACCEPT_LANGUAGE");
  BOOL finished = FALSE;
  while (accept_language && *accept_language && !finished)
    {
      char *dest;
      
      while (*accept_language && strchr(", '", *accept_language))
        {
          accept_language++;
        }

      for (dest=pLanguage; *accept_language && *accept_language!=',' && *accept_language!=';' && (dest-pLanguage)<buffer_size-1; )
        {
          *dest++= *accept_language++;
        }
      *dest=0;

      while (*accept_language && strchr(", '", *accept_language))
        {
          accept_language++;
        }

      if (*accept_language==';')
        {
          while (*accept_language && *accept_language!=',')
            {
              accept_language++;
            }
        }

      /* Test for language codes currently supported. */
      if (!strcasecmp(pLanguage, "de") ||
          !strcasecmp(pLanguage, "en") ||
          !strcasecmp(pLanguage, "nl") ||
          !strcasecmp(pLanguage, "fr"))
        {
          finished = TRUE;
        }
      else if (!strncmp(pLanguage, "de-", 3) ||
               !strncmp(pLanguage, "en-", 3) ||
               !strncmp(pLanguage, "nl-", 3) ||
               !strncmp(pLanguage, "fr-", 3))
        {
          // For example de-de or de-ch, or nl-be
          finished = TRUE;
          pLanguage[2]=0;
        }
    }

  if (!finished)
    {
      /* Use "en" by default */
      strncpy(pLanguage, "en", buffer_size);
    }

  return finished;
}


/* Set the `O_NONBLOCK' flag of DESC if VALUE is nonzero,
 * or clear the flag if VALUE is 0.
 *
 * @param desc    Filedescriptor to modify
 * @param value   1 Set O_NONBLOCK-Flag, 0 to reset NONBLOCK-Flag.
 *
 * @return 0 on success, or -1 on error with `errno' set.
 */
static
int
_set_nonblock_flag (int desc, int value)
{
  int oldflags;

  oldflags = fcntl (desc, F_GETFL, 0);

  /* If reading the flags failed, return error indication now. */
  if (oldflags == -1)
    return -1;

  /* Set just the flag we want to set. */
  if (value != 0)
    oldflags |= O_NONBLOCK;
  else
    oldflags &= ~O_NONBLOCK;

  /* Store modified flag word in the descriptor. */
  return fcntl (desc, F_SETFL, oldflags);
}



/**
 * Pipe a stringbuffer through a filter application
 *
 * 
 * @param sbio        Stringbuffer for input and post-processed output
 * @param argv        Argument vector
 *
 * @return TRUE if filter ran successfully
 */
static
BOOL
_sbfilter(STRINGBUFFER *sbio, char *const argv[])
{
#define READ_SIDE 0
#define WRITE_SIDE 1
  int tosed[2];
  int fromsed[2];
  BOOL result = FALSE;
  int status;

  STRINGBUFFER sbin;
  STRINGBUFFER sbout;
  memset(&sbout, 0, sizeof(sbout));
  memcpy(&sbin, sbio, sizeof(sbin));

  // Don't cancel application when writing to a closed pipe, happens if spawning of the child process failed.
  signal(SIGPIPE, SIG_IGN);

  if (!pipe(tosed) &&
      !pipe(fromsed))
    {
      int childpid=fork();
      if (childpid==0)
        {
          /* This is the child process */

          // Close unused pipe sides
          close( fromsed[ READ_SIDE ] );
          close( tosed[ WRITE_SIDE ] );

          // Set stdin/stdout
          dup2( fromsed[ WRITE_SIDE ], STDOUT_FILENO );
          dup2( tosed[ READ_SIDE ], STDIN_FILENO );
          close( fromsed[ WRITE_SIDE ]);
          close( tosed[ READ_SIDE ] );

          execvp(argv[0], argv);
          fprintf(stderr, "Starting child failed reason %s\n", strerror(errno));
          exit(EXIT_FAILURE);
        }
      else
        {
          /* This is the parent process */
          const char *src = sbget(sbio);
          int bytes_left = sblength(sbio);
          fd_set read_set;
          fd_set write_set;
          BOOL finished = FALSE;

          // Close unused pipe sides
          close( fromsed[ WRITE_SIDE ] );
          close( tosed[ READ_SIDE ] );

          _set_nonblock_flag( fromsed[ READ_SIDE ], TRUE );
          _set_nonblock_flag( tosed[ WRITE_SIDE ], TRUE );

          while (!finished)
            {
              FD_ZERO(&read_set);
              FD_ZERO(&write_set);
              FD_SET(fromsed[ READ_SIDE ], &read_set );
              if (bytes_left)
                {
                  FD_SET(tosed[ WRITE_SIDE ], &write_set );
                }

              if (select (FD_SETSIZE, &read_set, &write_set, NULL, NULL) > 0)
                {
                  if (FD_ISSET(fromsed[ READ_SIDE ], &read_set))
                    {
                      /* child has new processed data for us. */
                      char buffer[4096];
                      int bytes_read = read(fromsed[ READ_SIDE ], buffer, sizeof(buffer));
                      if (bytes_read > 0)
                        {
                          sbwrite(&sbout, buffer, bytes_read);
                        }
                      else
                        {
                          finished = TRUE;
                        }
                    }

                  if (bytes_left && FD_ISSET(tosed[ WRITE_SIDE ], &write_set))
                    {
                      /* child ready to accept new data on stdin */
                      int bytes_written = write(tosed[ WRITE_SIDE ], src, bytes_left);
                      if (bytes_written > 0)
                        {
                          src += bytes_written;
                          bytes_left -= bytes_written;

                          if (bytes_left==0)
                            {
                              /* Tell child that we have no additional data */
                              close(tosed[ WRITE_SIDE ]);
                              tosed[ WRITE_SIDE ]=-1;
                            }
                        }
                      else if (bytes_written == -1)
                        {
                          /* This happens if the child application could not be started. */
                          finished = TRUE;
                        }
                    }                  
                }
            }
        }

      if (sblength(&sbout))
        {
          memcpy(sbio, &sbout, sizeof(*sbio));
          sbclear(&sbin);
          result = TRUE;
        }
      else
        {
          sbclear(&sbout);
        }

      close(fromsed[ READ_SIDE ]);
      if (tosed[ WRITE_SIDE ]!=-1)
        {
          close(tosed[ WRITE_SIDE ]);
        }

      // Wait for the child to terminate to avoid zombie processes hanging around */
      waitpid(childpid, &status, 0);
    }

  return result;
}





/**
 * Post-Process HTML output.
 *
 * If a .sed script is defined, call sed using it as a filter.
 *
 * @param sbio        Stringbuffer for input and post-processed output
 * @param HlpFileName Filename of the help file,used to build the .sed file from
 */
void
PostProcess(STRINGBUFFER *sbio, const char *HlpFileName)
{
  char language[10];
  char sedfilename[500];
  char *argv[10];
  struct stat st;

  argv[0]="sed";
  argv[1]="-f";
  argv[2]=sedfilename;
  argv[3]=NULL;

  // Choose the language from the environment variable
  ChooseLanguage(language, sizeof(language));

  // helpfile.hlp.de.sed
  if (HlpFileName!=NULL)
    {
      snprintf(sedfilename, sizeof(sedfilename), "%s.%s.sed", HlpFileName, language);
    }
  if ( HlpFileName!=NULL && 0==stat(sedfilename, &st) )
    {
      _sbfilter(sbio, argv);
    }
  else
    {
      // helpfile.hlp.sed
      if (HlpFileName!=NULL)
        {
          snprintf(sedfilename, sizeof(sedfilename), "%s.sed", HlpFileName);
        }
      if ( HlpFileName!=NULL && 0==stat(sedfilename, &st) )
        {
          _sbfilter(sbio, argv);
        }
      else
        {
          // winhelpcgi.cgi.de.sed
          snprintf(sedfilename, sizeof(sedfilename), "%s.%s.sed", CGI_NAME, language);
          if ( 0==stat(sedfilename, &st) )
            {
              _sbfilter(sbio, argv);
            }
          else
            {
              // winhelpcgi.cgi.sed
              snprintf(sedfilename, sizeof(sedfilename), "%s.sed", CGI_NAME);
              if ( 0==stat(sedfilename, &st) )
                {
                  _sbfilter(sbio, argv);
                }
            }
        }
    }
}



/**
 * Perform gzip-compression on the STRINGBUFFER of the client understands it.
 *
 * @param sbio        Stringbuffer for input and post-processed output
 */
void
GzipEncode(STRINGBUFFER *sbio)
{
  const char *http_accept_encoding = getenv("HTTP_ACCEPT_ENCODING");
  char *argv[10];

  if (http_accept_encoding != NULL)
    {
      if (*http_accept_encoding=='\'')
        {
          http_accept_encoding++;
        }

      if (!memcmp(http_accept_encoding, "gzip", 4))
        {
          argv[0]="gzip";
          argv[1]="-1c";
          argv[2]=NULL;
          
          if (_sbfilter(sbio, argv))
            {
              fputs("Content-Encoding: gzip\n", stdout);
            }
        }
    }
}
