/*
 * Help Access Library 
 * A CGI-Program to read Windows Help files in a Web Browser.
 *
 * Copyright (C) 1995-2004 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* stringbuffer.c -- part of WinHelpCGI

   This piece maintains the simple STRINGBUFFER structures used
   by winhelpcgi similar to in-memory FILE *.
 
   (C)2004 Herd Software Development, Kettelerstr. 35 / 68642 Buerstadt / Germany"
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "stringbuffer.h"       // STRINGBUFFER Appendable strings structures

#define BLOCK_SIZE 0x4000



/**
 * Append the given number of bytes to the end of
 * the string buffer.
 *
 * @param sb  Pointer to the string-buffer
 * @param str Pointer to the string to be appended
 * @param len Number of bytes to be appended
 */
void
sbwrite(STRINGBUFFER *sb, const char *str, int len)
{
  if (len + sb->pos + 1 > sb->reserved)
    {
      sb->reserved += ((len + BLOCK_SIZE -1) / BLOCK_SIZE * BLOCK_SIZE);
      sb->string = (char *) realloc(sb->string, sb->reserved);
      if (!sb->string)
        {
          perror("sbwrite: realloc");
          exit( EXIT_FAILURE );
        }
    }
  
  memcpy(sb->string + sb->pos, str, len);
  sb->pos += len;
  sb->string[sb->pos]=0;
}


/**
 * Add a string to a string-buffer
 *
 * @param sb  Pointer to the string-buffer
 * @param str Pointer to the string to be appended
 *
 */
void
sbadd(STRINGBUFFER *sb, const char *str)
{
  sbwrite(sb, str, strlen(str));
}


/**
 * Clear a string buffer after usage
 *
 * @param sb  Pointer to the string-buffer
 */
void
sbclear(STRINGBUFFER *sb)
{
  if (sb->string)
    {
      free(sb->string);
    }
  sb->string=NULL;
  sb->reserved=
  sb->pos=0;
}


/**
 * Return a pointer to the stringbuffer contents
 *
 * @param sb     Pointer to the string-buffer
 * @param return a pointer to the string buffer contents 
 */
const char *
sbget(STRINGBUFFER *sb)
{
  return
    sb->string
    ? sb->string
    : "";
}


/**
 * printf-type formatting into a string-Buffer
 *
 *
 * @param sb     Pointer to the string-buffer
 * @param format sprintf format string.
 */
void
sbprintf(STRINGBUFFER *sb, const char *format, ...)
{
  char sbuffer[500];
  int retval;
  va_list ap;
  va_start (ap, format);

  retval = vsnprintf(sbuffer, sizeof(sbuffer), format, ap);
  if (retval > sizeof(sbuffer)-2)
    {
      char *dbuffer = (char *) malloc(retval + 2);
      vsnprintf(dbuffer, retval+1, format, ap);

      sbadd(sb, dbuffer);
      free(dbuffer);
    }
  else
    {
      sbadd(sb, sbuffer);
    }
  va_end (ap);
}



/**
 * Add a string to the string buffer, replacing special characters
 * by their HTML entities
 *
 * @param sb     Pointer to the string-buffer
 * @param src    Pointer to the source string to be added to the string buffer
 */
void
sbhtmlputs(STRINGBUFFER *sb, const char *src)
{ 
  for (;
       *src;
       ++src)
    {
      switch (((unsigned char) *src) & 0xff)
        {
        case 0xe4: sbadd(sb, "&auml;"); break;
        case 0xf6: sbadd(sb, "&ouml;"); break;
        case 0xfc: sbadd(sb, "&uuml;"); break;
        case 0xc4: sbadd(sb, "&Auml;"); break;
        case 0xd6: sbadd(sb, "&Ouml;"); break;
        case 0xdc: sbadd(sb, "&Uuml;"); break;
        case 0xdf: sbadd(sb, "&szlig;"); break;
        case '<' : sbadd(sb, "&lt;"); break;
        case '>' : sbadd(sb, "&gt;"); break;
        case '&' : sbadd(sb, "&amp;"); break;
        case '\"': sbadd(sb, "&quot;"); break;
        case 183:
          {
            /* A Bullet in "Symbol" font. &bull;=&#8226; looks better on browsers new enough to understand it,
             * But internet explorer chokes if &#8226; is encoded while symbol font is selected.
             */
            const char *user_agent = getenv("HTTP_USER_AGENT");
            
            if (NULL==user_agent ||
                NULL!=strstr(user_agent, "MSIE "))
              {
                sbadd(sb, "&#183;"); 
              }
            else
              {
                sbadd(sb, "&#8226;"); 
              }
          }
          break;
        case  222: sbadd(sb, "&rarr;"); break; // a double right arrow in Symbol font
        default:   if (*src & 0x80)
          {
            sbprintf(sb, "&#%d;", ((unsigned char) *src) & 0xff);
          }
        else
          if (*src >= ' ')
            {
              sbputchar(sb, *src);
            }
        }
    }
}


/**
 * Write the string buffer into the file and clear it.
 *
 * @param sb     Pointer to the string-buffer
 * @param fout   Pointer to the output file.
 */
void
sbflush(STRINGBUFFER *sb, FILE *fout)
{
  if (sb->pos)
    {
      fwrite(sb->string, 1, sb->pos, fout);
    }
  sbclear(sb);
}


/**
 * Add one character to a stringbuffer.
 *
 * @param sb     Pointer to the string-buffer
 * @param ch     The character to be added.
 */
void
sbputchar(STRINGBUFFER *sb, char ch)
{
  char *dst;
  if (sb->pos + 2 > sb->reserved)
    {
      sb->reserved += BLOCK_SIZE;
      sb->string = (char *) realloc(sb->string, sb->reserved);
    }

  dst = sb->string + (sb->pos++);
  *dst++=ch;
  *dst++=0;
}




/**
 * Append the contents of the source string buffer
 * to the end of the destination string buffer and
 * clear the source.
 *
 * @param sb     Pointer to the destination string-buffer
 * @param src    Pointer to the source string-buffer that will be deleted after copying
 */
void
sbcat(STRINGBUFFER *sb, STRINGBUFFER *src)
{
  sbwrite(sb, sbget(src), sblength(src));
  sbclear(src);
}
