/*
 * Help Access Library 
 * A CGI-Program to read Windows Help files in a Web Browser.
 *
 * Copyright (C) 1995-2004 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* virtchmfs.c -- part of WinHelpCGI

   Implements a virtual CHM sourcecode file system:

   (C)1996-2004 Herd Software Development, Kettelerstr. 35 / 68642 Buerstadt / Germany"
*/

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#define CREATE_HLPACCES_DLL
#include "hlpacces.h"
#include "top.h"
#include "stringbuffer.h"
#include "virtchmfs.h"
#include "winhelpcgi.h"         // Header-File to connect the pieces of WinHelpCGI application
#include "shgimpor.h"           // Import of Bitmaps in Help File Format
#include "dib2png.h"            // DIB Image -> PNG file conversion
#include "wmf2png.h"            // Metafile Image -> PNG file conversion

/**
 * Informations stored for an opened .HLP file interpreted
 * as a virtual CHM source code.
 */
struct vchm
{
  /// Internal Filesystem handle of the Help file obtained by IFSOpen
  HIFS   ifs;

  /// Handle of the |TOPIC internal file of the IFS, NULL as long as we only access images.
  HTOPIC topic;

  /// Filename of the Help file.
  char *HlpFileName;

};

static void _VCHMFormVirtualName(struct vchm *pchm, char separator, const char *extension, DWORD dwTopicOffset, char *virtname, int virtname_size);



/**
 * Open a virtual CHM File System object
 *
 * @param HelpFileName     Filename of the physical help file.
 *
 * @return struct vfs virtual filesystem object or NULL if opening the file was not successfull.
 */
struct vchm*
VCHMOpen(const char *HelpFileName)
{
  HIFS ifs;
  struct vchm *result = NULL;

  if (NULL!=(ifs = IFSOpen(HelpFileName)))
    {
      result = (struct vchm *) calloc(sizeof(struct vchm), 1);
      if (result)
        {
          result->ifs = ifs;
          result->HlpFileName = strdup(HelpFileName);
        }
      else
        {
          /* Out of memory=?! */
          IFSClose(ifs);
        }
    }

  return result;
}


/**
 * Close a virtual CHM File System object.
 *
 * @param pchm             Pointer to the virtual filesystem object opened by VCHMOpen
 */
void
VCHMClose(struct vchm *pchm)
{
  if (pchm->topic)
    {
      HlpTopicClose(pchm->topic);
    }

  free(pchm->HlpFileName);
  IFSClose(pchm->ifs);
  free(pchm);
}



/* Access Contents of an Image in image/png file format.
 *
 * Graphics are stored as Internal Files named BMnumber or |BMnumber in
 * the Internal File System. These files have the same structure as MRB and
 * SHG-Files created outside of the Help-Files, but there are two other
 * Compression schemes.
 *
 * The SHGImport routine takes a Memory-Block in this SHG-File format
 * and converts it to a DIB (Device Independend Bitmap) or a Metafile
 * depending on the contents of the File.
 *
 * The Segmentation Parameters Handle defined in the last Parameter of
 * SHGImport has the same structure defined in seg.h first introduced with
 * ThinHelp Version 1.0 and contains the hotspot rectangle and
 * Macro/Jump informations.
 *
 * Things get much more complex when an embedded Window DLL is to be
 * used by the Help File. We don't want to discuss this here...
 *
 * @param pchm         Virtual CHM filesystem opened by a call to VCHMOpen
 * @param sbout        Output STRINGBUFFER object
 * @param HlpImage     Image name such as |bm1
 * @param mimetype     On return contains the content-type
 * @param mimetypesize Size of mimetype
 *
 * @return TRUE on success
 */
static
BOOL
_VFSGetImage(struct vchm *pchm, STRINGBUFFER *sbout, const char *HlpImage, char *mimetype, int mimetypesize)
{
  BOOL result = FALSE;
  HIFSFILE  ifsfile;
  DWORD     dwSHGSize;
  HGLOBAL   hDIB=NULL;
  LPSTR     lpSHG;
  HMETAFILE hMetaFile=NULL;
  int       iImageSize;
  METAFILEHEADER metaFileHeader;
  
  /* Read the "bm0.." or "|bm0.." - File, which is always in SHG-File format */
  ifsfile = IFSOpenFile(pchm->ifs, HlpImage);
  if (ifsfile)
    {
      dwSHGSize = IFSSeekFile(ifsfile, 0, SEEK_END);
      IFSSeekFile(ifsfile, 0, SEEK_SET);
      
      lpSHG = (LPSTR) malloc(dwSHGSize);
      if (lpSHG)
        {
          IFSReadFile(ifsfile, lpSHG, dwSHGSize);
          
          /* Convert SHG-File to standard DIB or Metafile format */
          SHGImport(lpSHG,                          // Input Data Block
                    &hDIB,                          // DIB-Handle
                    &hMetaFile,                     // Metafile-Handle for SHG containing WMF
                    &iImageSize,                    // Will be filled with Informations on placeable Metafiles
                    &metaFileHeader,                // Placeable Metafile header Informations
                    NULL                            // Segmentation Informations
                    );
          
          /* Note: SHGs can consist either of Bitmaps or of
           *   Metafiles.
           */
          free(lpSHG);
          
          if (hDIB)
            {
              /* Convert the Standard format DIB to a standard PNG image */ 
              Dib2Png(sbout, hDIB);
              GlobalFree(hDIB);
            }
          else if (hMetaFile)
            {
#ifdef HAVE_LIBWMF
              Wmf2Png(sbout, hMetaFile, iImageSize, &metaFileHeader);
#endif
                  
#ifndef _Windows
              GlobalFree((HGLOBAL) hMetaFile);
#else
#error "Don't do that..."              
#endif
            }
          
          /* MIME content-type information */
          strncpy(mimetype, "image/png", mimetypesize);
          result = TRUE;
        }
      
      IFSCloseFile(ifsfile);
    }

  return result;
}

/**
 * Return the contents of a topic in HTML format.
 *
 * @param pchm                 Virtual CHM filesystem opened by a call to VCHMOpen
 * @param sbout                Output STRINGBUFFER object
 * @param HlpContextId         Context ID of the topic or NULL if no context ID available.
 * @param HlpTopicOffset       Topicoffset as 8-Digit hex tring or NULL if no TopicOffset available.
 * @param dwFormattingFlags    Additional formatting control flags
 * @param virtual_root         Name used in relative links for the root of the help file
 * @param mimetype             On return contains the content-type
 * @param mimetypesize         Size of mimetype
 *
 * @return TRUE on success
 */
static
BOOL
_VFSGetHtml(struct vchm *pchm, STRINGBUFFER *sbout, const char *HlpContextId, 
            const char *HlpTopicOffset, unsigned int dwFormattingFlags, const char *virtual_root,
            char *mimetype, int mimetypesize)
{
  BOOL result = FALSE;
  HTOPIC topic = VCHMGetTopicHandle(pchm);
  
  if (topic != NULL)
    {
      EnumerateTopic(sbout, pchm->ifs, topic, pchm->HlpFileName, HlpTopicOffset, HlpContextId, dwFormattingFlags, virtual_root);
      result= TRUE;
    }

  return result;
}



struct hhpinfo
{
  STRINGBUFFER *sbout;
  struct vchm *pchm;

  // Only used for .hhk
  HIFSFILE kwdata;
};


/**
 * Callback function called by EnumerateContents for every
 * topic Title found in the Help file.
 *
 * @param lpszTitle      Title-String ($-Footnote)
 * @param dwTopicOffset  Title location in the |TOPIC file.
 * @param lParam         Application defined parameter
 */
static
BOOL FAR PASCAL
HhpTitleEnumerationProcedure(LPCSTR lpszTitle, DWORD dwTopicOffset, LPARAM lParam)
{
  struct hhpinfo *info = (struct hhpinfo *) lParam;
  char virtname[500];

  _VCHMFormVirtualName(info->pchm, '/', ".html", dwTopicOffset, virtname, sizeof(virtname));

  sbprintf(info->sbout, "%s\n", virtname);
  
  return TRUE;
}


/**
 * Add an [alias] section to the .hhp file to associate
 * the names used in the [map] file to the filenames of the .html
 * files.
 *
 * @param pchm                 Virtual CHM filesystem opened by a call to VCHMOpen
 * @param sbout                Output STRINGBUFFER object
 */
static
void
_VFSAddAliasSection(struct vchm *pchm, STRINGBUFFER *sbout)
{
  HIFSFILE ctxomap;

  sbadd(sbout, "\n[ALIAS]\n");

  // Open the context map file (if it exists)
  ctxomap = IFSOpenFile(VCHMGetIfsHandle(pchm), "|CTXOMAP");
  if (ctxomap != NULL)
    {
      BYTE buffer[20];
      
      if (IFSReadFile(ctxomap, buffer, 2)==2)
        {
          int record_count = buffer[0] | buffer[1]<<8;
          int index;
          for (index=0; index<record_count; index++)
            {
              if (IFSReadFile(ctxomap, buffer, 8)==8)
                {
                  DWORD dwTopicOffset = buffer[4]|buffer[5]<<8|buffer[6]<<16|buffer[7]<<24;

                  char virtname_id[100];
                  char virtname_url[100];
                  _VCHMFormVirtualName(pchm, '_', "", dwTopicOffset, virtname_id, sizeof(virtname_id));
                  _VCHMFormVirtualName(pchm, '/', ".html",dwTopicOffset, virtname_url, sizeof(virtname_url));
                  sbprintf(sbout, "%s=%s\n", virtname_id, virtname_url);
                }
            }
        }

      IFSCloseFile(ctxomap);
    }

}




/**
 * Return HtmlHelp project file (.hhp)
 *
 * @param pchm                 Virtual CHM filesystem opened by a call to VCHMOpen
 * @param sbout                Output STRINGBUFFER object
 * @param mimetype             On return contains the content-type
 * @param mimetypesize         Size of mimetype
 *
 * @return TRUE on success
 */
static
BOOL
_VFSGetHhp(struct vchm *pchm, STRINGBUFFER *sbout,
            char *mimetype, int mimetypesize)
{
  BOOL result = TRUE;
  time_t now = time(NULL);
  struct hhpinfo info;
  LPHELPFILETITLE tit;
  HTOPIC topic = VCHMGetTopicHandle(pchm);

  info.sbout = sbout;
  info.pchm = pchm;

  sbprintf(sbout, "; Html Help project file automatically generated by winhelpcgi " VERSION " %s\n\n", asctime( localtime( &now )));

  if (topic != NULL)
    {
      sbadd(sbout, "[OPTIONS]\n");
      if (NULL!=(tit=HlpGetHelpTitle(pchm->ifs)))
        {
          const char *ContentsContextID = HlpTopicFindContextIdForTopicOffset(topic, tit->ContentsTopicOffset);
          
          sbprintf(sbout, "Title=%s\n", tit->HelpTitle ? tit->HelpTitle : "-");
          sbprintf(sbout, "Default topic=");
          
          if (ContentsContextID)
            {
              sbprintf(sbout, "id/%s.html", ContentsContextID);
            }
          else
            {
              sbprintf(sbout, "topicoffset/0x%08x.html", tit->ContentsTopicOffset);
            }
          
          HlpFreeHelpTitle(tit);
          sbputchar(sbout, '\n');
        }

      sbadd(sbout, "Contents File=winhelpcgi.hhc\n");
      sbadd(sbout, "Index file=winhelpcgi.hhk\n");
      sbadd(sbout, "\n");
      
      sbadd(sbout, "[FILES]\n");
      HlpEnumTitles(pchm->ifs, HhpTitleEnumerationProcedure, (LPARAM) &info);

      // Create a reference to our .h file with the [map] data
      sbadd(sbout, 
            "\n[MAP]\n"
            "#include \"winhelpcgi.h\"\n");

      // Add an [alias] section
      _VFSAddAliasSection(info.pchm, sbout);
      
      // Set the content-type
      strncpy(mimetype, "text/plain", mimetypesize);
    }


  return result;
}



/**
 * Output the contents of a struct contentsentry to teh STRINGBUFFER
 *
 * @param pchm               Virtual CHM filesystem opened by a call to VCHMOpen
 * @param sbout              Output STRINGBUFFER object
 * @param contents           Pointer to the struct contentsentry
 * @param recurselevel       Recursion level meant to help formatting
 */
static
void
_VFSDownContents(struct vchm *pchm, STRINGBUFFER *sbout, struct contentsentry *contents, int recurselevel)
{
  char indent[100];
  HTOPIC topic=VCHMGetTopicHandle(pchm);
  
  memset(indent, '\t', recurselevel);
  indent[recurselevel]=0;

  if (contents->title!=NULL)
    {
      sbprintf(sbout,
               "%s<li> <object type=\"text/sitemap\">\n"
               "%s    <param name=\"name\" value=\"%s\">\n",
               indent,
               indent, contents->title);

      if (contents->contextstring && contents->contextstring[0])
        {
          LPCSTR ContextId=contents->contextstring;
          DWORD ContextIdTopicOffset = HlpIsTopic(pchm->ifs, HlpGenerateHashValue(contents->contextstring));   // Context If can be found.
          DWORD TitleTopicOffset = ContextIdTopicOffset;
          if (ContextIdTopicOffset != 0xffffffff)
            {
              char virtname[500];
              HlpGetTopicTitle(pchm->ifs, ContextIdTopicOffset, NULL, 0, &TitleTopicOffset);
              
              if (ContextIdTopicOffset != TitleTopicOffset)
                {
                  /* This is a Jump to a sub-anchor of the topic. */
                  LPSTR TitleContextID = HlpTopicFindContextIdForTopicOffset(topic, TitleTopicOffset);
                  if (TitleContextID)
                    {
                      snprintf(virtname, sizeof(virtname),
                               "id/%s.html#%s",
                               UrlEscape(TitleContextID),
                               UrlEscape(ContextId));
                    }
                  else
                    {
                      snprintf(virtname, sizeof(virtname),
                               "topicoffset/0x%08x.html#%s",
                               TitleTopicOffset,
                               UrlEscape(ContextId));
                      
                    }
                }
              else
                {
                  snprintf(virtname, sizeof(virtname),
                           "id/%s.html",
                           UrlEscape(ContextId));
                }

              sbprintf(sbout, "%s    <param name=\"Local\" value=\"%s\">\n",
                       indent, virtname);
            }
        }

      sbprintf(sbout,
               "%s     </object>\n"
               "%s</li>\n",
               indent,
               indent);
    }
  else if (contents->subtopiccount != 0)
    {
      int i;
      sbprintf(sbout, "%s<ul>\n", indent);

      for (i=0; i<contents->subtopiccount; i++)
        {
          _VFSDownContents(pchm, sbout, contents->subtopics[i], recurselevel+1);
        }
      
      sbprintf(sbout, "%s</ul>\n", indent);
    }
}





/**
 * Callback called from HlpEnumTitles for every topic title found in the File.
 * Used from _VFSGetHhc if we have no informations from a .cnt file.
 *
 * @param lpszTitle      Title-String ($-Footnote)
 * @param dwTopicOffset  Title location in the |TOPIC file.
 * @param lParam         Application defined parameter
 */
static
BOOL CALLBACK
_VFSTitleEnumerationProcedure(LPCSTR lpszTitle, DWORD dwTopicOffset, LPARAM lParam)
{
  struct hhpinfo *info = (struct hhpinfo *) lParam;
  STRINGBUFFER *sbout = info->sbout;
  char virtname[500];

  _VCHMFormVirtualName(info->pchm, '/', ".html",dwTopicOffset, virtname, sizeof(virtname));

  sbprintf(sbout,
           "  <li> <object type=\"text/sitemap\">\n"
           "      <param name=\"name\" value=\"%s\">\n",
           lpszTitle);
  
  sbprintf(sbout, 
           "      <param name=\"Local\" value=\"%s\">\n",
           virtname);

  sbprintf(sbout,
           "      </object>\n"
           "  </li>\n");
  
  return TRUE;
}


/**
 * Return HtmlHelp contents file  (.hhc)
 *
 * @param pchm                 Virtual CHM filesystem opened by a call to VCHMOpen
 * @param sbout                Output STRINGBUFFER object
 * @param mimetype             On return contains the content-type
 * @param mimetypesize         Size of mimetype
 *
 * @return TRUE on success
 */
static
BOOL
_VFSGetHhc(struct vchm *pchm, STRINGBUFFER *sbout,
            char *mimetype, int mimetypesize)
{
  BOOL result = TRUE;
  struct hhpinfo info;
  HTOPIC topic = VCHMGetTopicHandle(pchm);

  info.sbout = sbout;
  info.pchm = pchm;

  if (topic != NULL)
    {
      sbadd(sbout,
            "<html>\n"
            " <head>\n"
            " <meta name=\"GENERATOR\" content=\"winhelpcgi " VERSION "\">\n"
            " <!-- Sitemap 1.0 -->\n"
            " </head>\n"
            "<body>\n"
            );

      if (topic->contents != NULL)
        {
          /* Have hierarchical informations from the .cnt files. */
          _VFSDownContents(pchm, sbout, topic->contents, 1);
        }
      else
        {
          /* Don't have .cnt files. Use a linear view. */
          sbadd(sbout, " <ul>\n");
          HlpEnumTitles(pchm->ifs, _VFSTitleEnumerationProcedure, (LPARAM) &info);
          sbadd(sbout, " </ul>\n");
        }

      sbadd(sbout,
            "\n</body>\n</html>\n");
    }


  return result;
}



/**
 * Try to form a virtual file name from a Topic offset.
 *
 * @param pchm               Virtual CHM filesystem opened by a call to VCHMOpen
 * @param separator          Separator character used
 * @param dwTopicOffset      Topic offset we need a name for
 * @param virtname           Pointer to a buffer that will contain the virtual name
 * @param virtname_size      Size of the buffer for the virtual name
 */
static
void
_VCHMFormVirtualName(struct vchm *pchm, char separator, const char *extension, DWORD dwTopicOffset, char *virtname, int virtname_size)
{
  DWORD TitleTopicOffset;
  LPSTR TitleContextID;

  HlpGetTopicTitle(pchm->ifs, dwTopicOffset, NULL, 0, &TitleTopicOffset);
              
  TitleContextID = HlpTopicFindContextIdForTopicOffset(VCHMGetTopicHandle(pchm), TitleTopicOffset);
  if (TitleContextID)
    {
      snprintf(virtname, virtname_size,
               "id%c%s%s",
               separator,
               UrlEscape(TitleContextID),
               extension);

      free(TitleContextID);
    }
  else
    {
      snprintf(virtname, virtname_size,
               "topicoffset%c0x%08x%s",
               separator,
               TitleTopicOffset,
               extension);
    }
}




/**
 * Callback function called for every keyword in the help file.
 *
 * @param lpszKeyWord        Keyword name found in |KWBTREE file
 * @param dwDataOffset       Offset of keyword data in |KWDATA file.
 * @param uDataCount         Number of topicoffsets known for this keyword
 * @param lParam             Pointer to application object
 */
static
BOOL CALLBACK
_VFSKeywordEnumProc(LPCSTR lpszKeyWord, DWORD dwDataOffset, UINT uDataCount, LPARAM lParam)
{
  struct hhpinfo *info = (struct hhpinfo *) lParam;

  DWORD dwTopicOffset;
  int i;
  BYTE abTopicOffset[4];
  char virtname[100];

  sbprintf(info->sbout,
           "  <li> <OBJECT type=\"text/sitemap\">\n"
           "         <param name=\"Name\" value=\"%s\">\n",
           lpszKeyWord
           );
  
  IFSSeekFile(info->kwdata, dwDataOffset, SEEK_SET);
  for (i=0; i<uDataCount; ++i)
    {
      IFSReadFile(info->kwdata, &abTopicOffset, sizeof(abTopicOffset));
      
      dwTopicOffset = abTopicOffset[0]+(abTopicOffset[1]<<8)+(abTopicOffset[2]<<16)+(abTopicOffset[3]<<24);

      _VCHMFormVirtualName(info->pchm, '/', ".html", dwTopicOffset, virtname, sizeof(virtname));
      
      sbprintf(info->sbout,
               "         <param name=\"Local\" value=\"%s\">\n",
               virtname
               );
    }

  sbadd(info->sbout,
        "       </object>\n"
        "  </li>\n");
  
  return TRUE;
}



/**
 * Return HtmlHelp keyword file (.hhk)
 *
 * @param pchm                 Virtual CHM filesystem opened by a call to VCHMOpen
 * @param sbout                Output STRINGBUFFER object
 * @param mimetype             On return contains the content-type
 * @param mimetypesize         Size of mimetype
 *
 * @return TRUE on success
 */
static
BOOL
_VFSGetHhk(struct vchm *pchm, STRINGBUFFER *sbout,
            char *mimetype, int mimetypesize)
{
  BOOL result = TRUE;
  struct hhpinfo info;
  HTOPIC topic;

  info.sbout = sbout;
  info.pchm = pchm;
  info.kwdata = IFSOpenFile(pchm->ifs, "|KWDATA");

  if (NULL != (topic = VCHMGetTopicHandle(pchm)) &&
      topic->contents != NULL &&
      info.kwdata != NULL)
    {
      sbadd(sbout,
            "<html>\n"
            " <head>\n"
            " <meta name=\"GENERATOR\" content=\"winhelpcgi " VERSION "\">\n"
            " <!-- Sitemap 1.0 -->\n"
            " </head>\n"
            "<body>\n"
            );

      sbadd(sbout,
            "<ul>\n");

      HlpEnumKeyWords(pchm->ifs, _VFSKeywordEnumProc, "", "|KWBTREE", (LPARAM) &info);

      sbadd(sbout,
            "</ul>\n");

      sbadd(sbout,
            "\n</body>\n</html>\n");

      IFSCloseFile(info.kwdata);
    }

  return result;
}



/**
 * Return HtmlHelp .h include file with informations to go to the [map] section
 *
 * @param pchm                 Virtual CHM filesystem opened by a call to VCHMOpen
 * @param sbout                Output STRINGBUFFER object
 * @param mimetype             On return contains the content-type
 * @param mimetypesize         Size of mimetype
 *
 * @todo Does not support .HLP files compiled with the old HC30.EXE compiler.
 *
 * @return TRUE on success
 */
static
BOOL
_VFSGetH(struct vchm *pchm, STRINGBUFFER *sbout,
         char *mimetype, int mimetypesize)
{
  HIFSFILE ctxomap;
  
  // Open the context map file (if it exists)
  ctxomap = IFSOpenFile(VCHMGetIfsHandle(pchm), "|CTXOMAP");
  if (ctxomap != NULL)
    {
      BYTE buffer[20];

      if (IFSReadFile(ctxomap, buffer, 2)==2)
        {
          int record_count = buffer[0] | buffer[1]<<8;
          int index;
          for (index=0; index<record_count; index++)
            {
              if (IFSReadFile(ctxomap, buffer, 8)==8)
                {
                  DWORD dwMapID = buffer[0]|buffer[1]<<8|buffer[2]<<16|buffer[3]<<24;
                  DWORD dwTopicOffset = buffer[4]|buffer[5]<<8|buffer[6]<<16|buffer[7]<<24;

                  char virtname[100];
                  _VCHMFormVirtualName(pchm, '_', "", dwTopicOffset, virtname, sizeof(virtname));
                  sbprintf(sbout, "#define\t%s\t%d\n", virtname, dwMapID);
                }
            }
        }

      IFSCloseFile(ctxomap);
    }

  // Set the content-type
  strncpy(mimetype, "text/plain", mimetypesize);

  return TRUE;
}





/**
 * Return the handle of the |TOPIC file opened for this virtual CHM filesystem instance.
 * Opens the |TOPIC file if it has not been opened before.
 *
 * The handle must not be closed by the application.
 * It will be closed automatically on VCHMClose(...)
 *
 * @param pchm               Virtual CHM filesystem opened by a call to VCHMOpen
 *
 * @return Handle to the Topic file, NULL if opening the topic file failed.
 */
HTOPIC
VCHMGetTopicHandle(struct vchm *pchm)
{
  if (pchm->topic == NULL)
    {
      /* Topic file not opened yet. Open it now. */
      pchm->topic = HlpTopicOpen(pchm->ifs, HTO_FONTS | HTO_CONTEXT | HTO_KEYWORDS);
    }
  
  return pchm->topic;
}



/**
 * Return the handle of the internal filesystem of the Help file.
 *
 * The handle must not be closed by the application.
 * It will be closed automatically on VCHMClose(...)
 *
 * @param pchm               Virtual CHM filesystem opened by a call to VCHMOpen
 *
 * @return Handle to the Internal filesystem
 */
HIFS
VCHMGetIfsHandle(struct vchm *pchm)
{
  return pchm->ifs;
}


/**
 * Return the filename of the helpfile used to open the file.
 *
 * @param pchm               Virtual CHM filesystem opened by a call to VCHMOpen
 *
 * @return const Pointer to a buffer congtaining the filename
 */
const char *
VCHMGetHlpFileName(struct vchm *pchm)
{
  return pchm->HlpFileName;
}





/**
 * Fills a STINGBUFFER structure with informations matching
 * the requested filename within the virtual namespace defined
 * for .HLP files.
 *
 * @param pchm                 Virtual CHM filesystem opened by a call to VCHMOpen
 * @param sbout                Output STRINGBUFFER object
 * @param virtname             Filename of the virtual file
 * @param dwFormattingFlags    Additional formatting control flags
 * @param mimetype             On return contains the content-type
 * @param mimetypesize         Size of mimetype
 *
 * Naming scheme is:
 * id/contextid.html                 For regular file
 * topicoffset/0x12345678.html       For regular file by TopicOffset (If no context ID defined for topic)
 *
 * popup/id/contextid.html           For file in popip formatting (No navigation)
 * pupup/topicoffset/0x12345678.html
 * 
 * print/id/contextid.html           For file in print formatting (No navigation)
 * print/topicoffset/0x12345678.html
 *
 * png/bmxxx.png                     For an image file.
 *
 * @return TRUE on success
 */
BOOL
VCHMGetFile(struct vchm *pchm, STRINGBUFFER *sbout, const char *virtname, unsigned int dwFormattingFlags, char *mimetype, int mimetypesize)
{
  const char* HlpTopicOffset=NULL;
  const char* HlpImage=NULL;
  const char* HlpContextId=NULL;
  const char* HlpHhp=NULL;
  const char* HlpHhc=NULL;
  const char* HlpHhk=NULL;
  const char* HlpH=NULL;
  int len = strlen(virtname);
  char virtual_root[100];
  int virtual_root_len;

  strncpy(mimetype, "text/html", mimetypesize);
  virtual_root[0]=0;

  /* Separate virtual filename into relevant components */
  if (len>5 && !strcmp(virtname+len-5, ".html"))
    {
      BOOL isContextID=TRUE;
      
      if (!strncmp(virtname, "popup/", 6))
        {
          virtname += 6;
          dwFormattingFlags=HTMLTOPIC_POPUP | HTMLTOPIC_MARKPOPUP;
          strcat(virtual_root, "../");
        }
      else if (!strncmp(virtname, "print/", 6))
        {
          virtname += 6;
          dwFormattingFlags=HTMLTOPIC_PRINT;
          strcat(virtual_root, "../");
        }

      if (!strncmp(virtname, "id/", 3))
        {
          virtname += 3;
          strcat(virtual_root, "../");
        }
      else if (!strncmp(virtname, "topicoffset/0x", 14))
        {
          virtname += 12;
          isContextID = FALSE;
          strcat(virtual_root, "../");
        }

      {
        char *tHlpContextId;
        int contextid_length = strlen(virtname)-5;
        tHlpContextId = (char *) malloc( contextid_length  + 1);
        if (tHlpContextId)
          {
            strncpy(tHlpContextId, virtname, contextid_length );
            tHlpContextId[ contextid_length ] = 0;
            if (isContextID)
              {
                HlpContextId = tHlpContextId;
              }
            else
              {
                HlpTopicOffset = tHlpContextId;
              }
          }
      }
    }

  else if (len>4 && 
           !strcmp(virtname+len-4, ".png") && 
           !strncmp(virtname, "png/", 4)
           )
    {
      const char *p = virtname + 4;
      char *tHlpImage;
      int image_length = virtname+len-4 - p +1;

      strcat(virtual_root, "../");

      tHlpImage = (char *) malloc( image_length  + 1);
      if (tHlpImage)
        {
          tHlpImage[0]='|';
          
          strncpy(tHlpImage+1, p, image_length-1 );
          tHlpImage[ image_length ] = 0;
          HlpImage = tHlpImage;
        }
    }
  else if (len>4 &&
           !strcmp(virtname+len-4, ".hhp"))
    {
      HlpHhp=virtname;
    }
  else if (len>4 &&
           !strcmp(virtname+len-4, ".hhc"))
    {
      HlpHhc=virtname;
    }
  else if (len>4 &&
           !strcmp(virtname+len-4, ".hhk"))
    {
      HlpHhk=virtname;
    }
  else if (len>2 &&
           !strcmp(virtname+len-2, ".h"))
    {
      HlpH=virtname;
    }
  
  // Remove trailing / from virtual root
  virtual_root_len = strlen( virtual_root );
  if (virtual_root_len &&
      virtual_root[ virtual_root_len-1 ]=='/')
    {
      virtual_root[ --virtual_root_len ]=0;
    }

  return
    HlpImage!=NULL
    ? _VFSGetImage(pchm, sbout, HlpImage, mimetype, mimetypesize)
    : (HlpHhp!=NULL
       ? _VFSGetHhp(pchm, sbout, mimetype, mimetypesize)
       : (HlpHhc!=NULL
          ? _VFSGetHhc(pchm, sbout, mimetype, mimetypesize)
          : (HlpHhk!=NULL
             ? _VFSGetHhk(pchm, sbout, mimetype, mimetypesize)
             : (HlpH!=NULL
                ? _VFSGetH(pchm, sbout, mimetype, mimetypesize)
                : _VFSGetHtml(pchm, sbout, HlpContextId, HlpTopicOffset, dwFormattingFlags, virtual_root,
                              mimetype, mimetypesize)
                )
             )
          )
       );
}



/**
 * Return the timestamp when the given help file was last modified.
 *
 * @param pchm                 Virtual CHM filesystem opened by a call to VCHMOpen
 *
 * @return The internal compilation date of the .hlp file as a Unix-Style time-stamp.
 */
time_t
VCHMGetGenDate(struct vchm *pchm)
{
  time_t modified;
  struct stat st;

  LPHELPFILETITLE lpHelpTitle = HlpGetHelpTitle(pchm->ifs);
  if (lpHelpTitle)
    {
      /* Use the time when the help file has been compiled as the last-modified timestamp for the HTML code */
      modified = lpHelpTitle->GenDate;

      HlpFreeHelpTitle(lpHelpTitle);
    }
  else if (0==stat(pchm->HlpFileName, &st))
    {
      modified = st.st_mtime;
    }
  else
    {
      modified=time(NULL);
    }

  return modified;
}
