/*
 * Help Access Library 
 * A CGI-Program to read Windows Help files in a Web Browser.
 *
 * Copyright (C) 1995-2004 Bernd Herd, http://www.herdsoft.com
 *                                                                  
 * This program is free software; you can redistribute it and/or    
 * modify it under the terms of the GNU General Public License as   
 * published by the Free Software Foundation; either version 2 of   
 * the License.              
 *                                                                  
 * This program is distributed in the hope that it will be useful,  
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
 * GNU General Public License for more details.                     
 *                                                                  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software      
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        
 */
/* winhelpcgi.c -- part of WinHelpCGI

   CGI program used to display Windows Help Files in a web browser
   based on Help Access Library, generates HTML-Output for a Topic as a CGI-Program

   The RTF-Parser contained herein is limited to the subset of RTF-Coding
   applied by Help Access Library. It cannot be used to Convert fully
   featured RTF-Files from Word processing Applications into HTML format.

   (C)1996-2004 Herd Software Development, Kettelerstr. 35 / 68642 Buerstadt / Germany"
*/


#include <windows.h>
#include <fcntl.h>
#if defined(__WIN32__)
#  include <io.h>
#else
#  include <dirent.h>
#endif
#include <errno.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include <time.h>
#ifdef HAVE_LANGINFO_H
#include <langinfo.h>
#endif
#include <ctype.h>
#include "hlpacces.h"           // Help Access Library Functions
#include "winhelpcgi.h"         // Header-File to connect the pieces of WinHelpCGI application
#include "virtchmfs.h"          // Virtual CHM filesystem

#ifdef ENABLE_NLS
#include <libintl.h>

#define _(String) gettext (String)
#define gettext_noop(String) (String)
#define N_(String) gettext_noop (String)
#else
#define _(String) (String)
#endif

/* GetQueryVariable --------------------------------------------*
 * Get a Data-Block signifieing the contents of the Form query
 * variable. returns NULL if Variable nit defined.
 *--------------------------------------------------------------*/
LPCSTR GetQueryVariable(LPCSTR QueryString, LPCSTR VarName)
{
  int    VarNameLength= strlen(VarName);
  char * VarContents  = NULL;
  char * dst;
  
  int    cint;
  char   c;
  
  /*---------------- Scan for "KEYWORDS"-Variable ------------------*/
  while (QueryString &&
			memcmp(QueryString, VarName, strlen(VarName))
			)
    {
		if (NULL!=(QueryString = strchr(QueryString, '&')))
		  QueryString++;
    }
  
  /*---------------- If Variable has been found, return contents ---*/
  if (QueryString)
    {  
		VarContents = dst = (LPSTR) malloc(strlen(QueryString));
		
		for (QueryString+=VarNameLength;              /* Step over "KEYWORDS="... */
			  *QueryString&&*QueryString!='&';
			  )
		  {
			 switch (c=(*QueryString++))
				{ 
				case '+': 
				  c=' '; 
				  break;
				case '%': 
				  sscanf(QueryString, "%02x", &cint);
				  c=(char) cint;
				  QueryString+=2;
				}
			 
			 *dst++=c;
		  }
		
		*dst=0;
    }
  
  return VarContents;
}


/**
 * Return a string representation of the given timestamp that is
 * suitable for a last-modified timestamp.
 *
 * @param when     time given as a unix-type "time_t"
 *
 * @return Pointer to a static buffer with the string
 */
static
const char *
_TimeString(time_t when)
{
  static char last_modified[60];
  char *oldlocale = setlocale(LC_TIME, "C");
  strftime(last_modified, sizeof(last_modified),
           "%a, %d %h %Y %T GMT",
           localtime( &when ));
  
  if (oldlocale)
    {
      setlocale(LC_TIME, oldlocale);
    }
  
  return last_modified;
}


/**
 * Check if the context ID used in the virtual filename is created
 * Without the addtional knowledge of the context IDs found from the .cnt file.
 * This helps us to avoid google and other index machines having several copies
 * of the same contents.
 *
 * @param sbout                Output STRINGBUFFER to receive the name of the new location
 * @param pchm                 Virtual CHM filesystem opened by a call to VCHMOpen
 * @param virtname             Filename of the virtual file
 * @param mimetype             Buffer to receive the mimetype "hack/redirect" if it is a redirect
 * @param mimetypesize         Size of the mimetype.
 *
 * @return TRUE if this is indeed a redirect
 */
static
BOOL
_IsRedirect(STRINGBUFFER *sbout, struct vchm *pchm, const char *virtname, char *mimetype, int mimetypesize)
{
  int slash_len=strlen(virtname);
  const char *slash;
  BOOL Redirected = FALSE;
  
  /* The context ID is the portion of the filename between the last / and the .html */
  if (NULL!=(slash=strrchr(virtname, '/')) &&
      6   <(slash_len=strlen(slash)) &&
      slash_len < 16 &&
      !strcasecmp(slash+slash_len-5, ".html"))
    {
      char *context_id=strdup(slash+1);
      if (context_id != NULL)
        {
          context_id[slash_len-6]=0;

          {
            /* Trigger a redirect by "Location:..." if the Context-ID used to open the
             * file is not the context ID from the .cnt file but an auto-generated context id
             */
            DWORD ContextIdTopicOffset = HlpIsTopic(VCHMGetIfsHandle(pchm), HlpGenerateHashValue(context_id));
            if (ContextIdTopicOffset != 0xffffffff)
              {
                LPSTR ReverseContextID = HlpTopicFindContextIdForTopicOffset(VCHMGetTopicHandle(pchm), ContextIdTopicOffset);
                if (ReverseContextID)
                  {
                    if (strcasecmp(ReverseContextID, context_id))
                      {
                        strncpy(mimetype, "hack/redirect", mimetypesize);
                        Redirected=TRUE;
                        sbprintf(sbout, "location: %s.html\n\n", ReverseContextID);
                      }
                    
                    free(ReverseContextID);
                  }
              }
          }

          free(context_id);
        }
    }
  return Redirected;
}





/**
 * Forward a virtual file as a CGI answer.
 *
 * @param sbout                Output STRINGBUFFER
 * @param pchm                 Virtual CHM filesystem opened by a call to VCHMOpen
 * @param virtname             Filename of the virtual file
 * @param mimetype             On return contains the content-type
 * @param mimetypesize         Size of mimetype
 */
static
void
_VirtualAnswer(STRINGBUFFER *sbout, struct vchm *pchm, const char *virtname, char *mimetype, int mimetypesize)
{
  if (!_IsRedirect(sbout, pchm, virtname, mimetype, mimetypesize) &&
      !VCHMGetFile(pchm, sbout, virtname, HTMLTOPIC_NAVIGATION | HTMLTOPIC_MARKPOPUP | HTMLTOPIC_ADDHLPTITLE, mimetype, mimetypesize))
    {
      sbprintf(sbout,
               _("Can't open virtual file \"%s\". Possible reason: %s"),
               virtname, strerror(errno) );
    }
}




/* ================ Enumerate Help file Titles ======================= */

#define TITLES_PER_PAGE   500

typedef struct
{ 
  const char *HlpTitleScan;
  const char *HlpFileName;
  int         TitleCounter;
  int         Page;
  char        BaseUrl[500];
  int         PageCount;
  HTOPIC      htopic;
  const char *script_name;
  STRINGBUFFER *sbout;
} TITLEDATA, FAR *LPTITLEDATA;



/* TitleEnumerationProcedure
 *
 * Callback-Routine called by _EnumerateContents in call to HlpEnumTitles
 */
BOOL FAR PASCAL TitleEnumerationProcedure(LPCSTR lpszTitle, DWORD dwTopicOffset, LPARAM lParam)
{
  LPTITLEDATA lptit = (LPTITLEDATA) lParam;
  const char *HlpTitleScan = lptit->HlpTitleScan;
  const char *p=NULL;
  char *      dup;
  STRINGBUFFER *sbout = lptit->sbout;
  
  if (!HlpTitleScan || !*HlpTitleScan ||
		NULL!=(p=strstr(lpszTitle, HlpTitleScan)))
	 {
		lptit->TitleCounter++;
		
      if (lptit->TitleCounter >= TITLES_PER_PAGE * lptit->Page)
        {
          LPCSTR contextID = HlpTopicFindContextIdForTopicOffset(lptit->htopic, dwTopicOffset);
          if (contextID)
            {
              sbprintf(sbout, "<LI><a href=\"%s/%s/id/%s.html\">", lptit->script_name, lptit->HlpFileName, UrlEscape(contextID));
            }
          else
            {
              sbprintf(sbout, "<LI><a href=\"%s/%s/topicoffset/0x%08x.html\">", lptit->script_name, lptit->HlpFileName, dwTopicOffset);
            }
          
          if (p)
            {
              dup=strdup(lpszTitle);
              dup[p-lpszTitle]=0;
              sbhtmlputs(sbout, dup);
              free(dup);
              sbprintf(sbout, "<b>");
              dup=strdup(p);
              dup[strlen(HlpTitleScan)]=0;
              sbhtmlputs(sbout, dup);
              free(dup);
              sbprintf(sbout, "</b>");
              sbhtmlputs(sbout, p+strlen(HlpTitleScan));
            }
          else
            {
              sbhtmlputs(sbout, *lpszTitle ? lpszTitle : _("<No Title - maybe popup...>"));
            }
          
          
          sbprintf(sbout, "</a>\n");
        }
	 }
  
  return lptit->TitleCounter < (TITLES_PER_PAGE * (lptit->Page+1) );
}


typedef struct
{
  HIFS        ifs;
  HIFSFILE    hKWData;
  LPCSTR      HlpFileName;
  int         TopicCounter;
  int         Page;
  HTOPIC      htopic;
  const char *script_name;
  STRINGBUFFER *sbout;
} TKWLIST, FAR*LPTKWLIST;


static
BOOL FAR PASCAL
KWEnumProc(LPCSTR lpszKeyWord, DWORD dwDataOffset, UINT uDataCount, LPARAM lParam)
{
  LPTKWLIST tKWList=(LPTKWLIST)lParam;
  STRINGBUFFER *sbout = tKWList->sbout;
  int       i;
  DWORD     dwTopicOffset, dwTopicOffsetStart;
  BYTE      abTopicOffset[4];
  char      szTopicTitle[1024];
  
  if (tKWList->TopicCounter >= TITLES_PER_PAGE * tKWList->Page)
    {
      IFSSeekFile(tKWList->hKWData, dwDataOffset, SEEK_SET);
      
      sbadd(sbout, "<dt>");
      sbhtmlputs(sbout, lpszKeyWord);
      sbadd(sbout, "</dt>\n");
      
      for (i=0; i<uDataCount; ++i)
        {
          IFSReadFile(tKWList->hKWData, &abTopicOffset, sizeof(abTopicOffset));
          
          dwTopicOffset = abTopicOffset[0]+(abTopicOffset[1]<<8)+(abTopicOffset[2]<<16)+(abTopicOffset[3]<<24);
          
          if (HlpGetTopicTitle(tKWList->ifs, dwTopicOffset, szTopicTitle, sizeof(szTopicTitle), &dwTopicOffsetStart))
            {
              LPCSTR contextID = HlpTopicFindContextIdForTopicOffset(tKWList->htopic, dwTopicOffset);
              if (contextID)
                {
                  sbprintf(sbout, "<dd><a href=\"%s/%s/id/%s.html\">",
                         tKWList->script_name,
                         tKWList->HlpFileName,
                         UrlEscape(contextID));
                }
              else
                {
                  sbprintf(sbout, "<dd><a href=\"%s/%s/topicoffset/0x%08x.html\">",
                         tKWList->script_name,
                         tKWList->HlpFileName,
                         dwTopicOffset);
                }
              sbhtmlputs(sbout, szTopicTitle);
              sbadd(sbout, "</a></dd>\n");
            }
        }
    }

  return tKWList->TopicCounter++<(tKWList->Page+1) * TITLES_PER_PAGE;
}


/* DisplayKeyWordListing
 *
 * Start enumeration of the contents of a help file.
 *
 * @param sbout           Output STRINGBUFFER
 */
static
void
DisplayKeyWordListing(STRINGBUFFER *sbout, HIFS ifs, LPTITLEDATA tit, LPCSTR lpcszKeyWord)
{
  TKWLIST tKWList;
  
  sbadd(sbout, "<hr>\n<h2>Keyword Listing</h2>\n<dl>\n");
  
  tKWList.hKWData     = IFSOpenFile(ifs, "|KWDATA");
  tKWList.ifs         = ifs;
  tKWList.HlpFileName = tit->HlpFileName;
  tKWList.TopicCounter= 0;
  tKWList.Page        = tit->Page;
  tKWList.htopic      = HlpTopicOpen(ifs, /*HTO_FONTS |*/ HTO_CONTEXT/* | HTO_KEYWORDS*/);
  tKWList.script_name = getenv("SCRIPT_NAME");
  tKWList.sbout       = sbout;
  
  if (tKWList.hKWData!=NULL)
	 {
      HlpEnumKeyWords(ifs, KWEnumProc, lpcszKeyWord, "|KWBTREE", (LPARAM)(LPSTR)&tKWList);
      IFSCloseFile(tKWList.hKWData);
	 }
  else
	 sbadd(sbout, "Internal File KWDATA not found\n");
  
  HlpTopicClose(tKWList.htopic);

  /* -------- Write a Trailer ------------------------*/
  sbadd(sbout, "</dl>\n<hr>");
} // DisplayKeyWordListing



/**
 * Render the upper part of a contents view file with
 * Informations about the file and the form elements for
 * entering search criteria
 *
 * @param sbout           Output STRINGBUFFER
 * @param HlpFileName     Filename of the Help file
 * @param ifs             Internal Filesystem handle of the .HLP file
 * @param lpHelpTitle     Helpfile system informations retrieved
 * @param HlpTitleScan    Search string to find in help titles
 * @param HlpKeyWord      Search string to find in keywords
 */
static
void
_ContentsHeading(STRINGBUFFER *sbout, const char *HlpFileName, HIFS ifs, LPHELPFILETITLE lpHelpTitle, const char *HlpTitleScan, const char *HlpKeyWord)
{
  const char *script_name = getenv("SCRIPT_NAME");

  sbadd(sbout,
        "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
        "<html>\n<head>\n"
        "<title>");
  sbhtmlputs(sbout, lpHelpTitle->HelpTitle ? lpHelpTitle->HelpTitle : HlpFileName);
  
  sbadd(sbout, "</title>\n");
  
  sbprintf(sbout, "<link rel=\"up\" href=\"%s\" title=\"%s\">\n", script_name, _("Contents"));
  sbprintf(sbout, "<link rel=\"bookmark\" href=\"http://www.herdsoft.com/\" title=\"Herd Software Development\">\n");
  sbprintf(sbout, "<link rel=\"bookmark\" href=\"http://www.herdsoft.com/catalog/ehlp2rtf.html\" title=\"Help to RTF Converter\">\n");
  sbprintf(sbout, "<link rel=\"bookmark\" href=\"http://www.herdsoft.com/linux/produkte/winhelpcgi.html\" title=\"WinHelpCGI project page\">\n");
  sbprintf(sbout, "<link rel=\"search\" href=\"%s/%s/?KEYWORD=\" title=\"Glossary\">\n", script_name, HlpFileName);
}

/**
 * Render the upper part of a contents view file with
 * Informations about the file and the form elements for
 * entering search criteria
 *
 * @param sbout           Output STRINGBUFFER
 * @param HlpFileName     Filename of the Help file
 * @param ifs             Internal Filesystem handle of the .HLP file
 * @param lpHelpTitle     Helpfile system informations retrieved
 * @param HlpTitleScan    Search string to find in help titles
 * @param HlpKeyWord      Search string to find in keywords
 */
static
void
_ContentsStart(STRINGBUFFER *sbout, const char *HlpFileName, HIFS ifs, LPHELPFILETITLE lpHelpTitle, const char *HlpTitleScan, const char *HlpKeyWord)
{
  const char *script_name = getenv("SCRIPT_NAME");

  sbadd(sbout, "</head>\n");
  sbadd(sbout, "<h1>");
  sbhtmlputs(sbout, lpHelpTitle->HelpTitle ? lpHelpTitle->HelpTitle : HlpFileName);
  sbprintf(sbout, "</h1>\n"
         "<h2>%s</h2>",
         _("File Informations"));
  
  /* -------- Display some Help File informations ----*/
  sbprintf(sbout, "<table border=\"1\">"
         "<tr><td><b>%s:</b></td><td>",
         _("Filename"));
  sbhtmlputs(sbout, HlpFileName);
  sbadd(sbout, "</td></tr>\n");
  
  if (lpHelpTitle->HelpTitle)
    {
      sbprintf(sbout, "<tr><td><b>%s:</b></td><td>", _("File Title"));
      sbhtmlputs(sbout, lpHelpTitle->HelpTitle);
      sbadd(sbout, "</td></tr>\n");
    }
  
  if (lpHelpTitle->Copyright)
    {
      sbprintf(sbout, "<tr><td><b>%s:</b></td><td>", _("Copyright"));
      sbhtmlputs(sbout, lpHelpTitle->Copyright);
      sbadd(sbout, "</td></tr>\n");
    }
  
  if (lpHelpTitle->Citation)
    {
      puts("<tr><td><b>Citation:</b></td><td>");
      sbhtmlputs(sbout, lpHelpTitle->Citation);
      sbadd(sbout, "</td></tr>\n");
    }
  
  {
    char buffer[500];
#ifdef HAVE_LANGINFO_H
    strftime(buffer, sizeof(buffer), nl_langinfo(D_T_FMT), localtime(&lpHelpTitle->GenDate));
#else
    strncpy(buffer, asctime ( localtime( &lpHelpTitle->GenDate) ) , sizeof(buffer) );
#endif
    
    sbprintf(sbout, "<tr><td><b>%s</b></td><td>%s</td></tr>",
           _("Date compiled"),
           buffer);
  }
  
  sbprintf(sbout, "<tr><td><b>%s</b></td><td>%d</td></tr>",
         _("Number of Topics"),
         HlpGetTopicCount(ifs));
  
  sbprintf(sbout, "<tr><td><b>%s</b></td><td>%d</td></tr>",
         _("Number of Keywords"),
         HlpGetKeywordCount(ifs));                 
  
  if (lpHelpTitle->Citation)
    {
      
      sbadd(sbout, "<tr><td><b>Citation:</b></td><td>");
      sbhtmlputs(sbout, lpHelpTitle->Citation);
      sbadd(sbout, "</td></tr>\n");
    }
  
  
  sbadd(sbout, "</table>");
  
  /* -------- Allow entering search keys -------------*/
  sbprintf(sbout, "<hr><h2 id=\"hideprint\">%s</h2>"
         "<table border=\"0\" id=\"hideprint\">\n"
         "<tr><th>%s</th>\n"
         "<td>"
         "<form method=get action=\"%s/%s/\">"
         "<input name=\"TITLESCAN\" size=\"50\" value=\"%s\">"
         "<INPUT TYPE=SUBMIT value=\"%s\">"
         "</form>"
         "</td>\n"
         "</tr>\n",
         _("Search functions"),
         _("Text to find in title:"),
         script_name,
         HlpFileName,
         HlpTitleScan?HlpTitleScan:"",
         _("Start Search")
         );
  
  /* -------- Allow entering search keys -------------*/
  sbprintf(sbout, "<tr><th>%s</th>\n"
         "<td>\n"
         "<form method=get action=\"%s/%s/\">"
         "<input name=\"KEYWORD\" size=50 value=\"%s\">"
         "<INPUT TYPE=SUBMIT value=\"%s\">"
         "</form>"
         "</td>\n"
         "</tr>\n"
         "</table>\n",
         _("List starting with search KeyWord:"),
         script_name,
         HlpFileName,
         HlpKeyWord?HlpKeyWord:"",
         _("Start Search")
         );
}


/**
 * Render the lower part of a contents view file with
 * Informations about the file and the form elements for
 * entering search criteria
 *
 * @param sbout           Output STRINGBUFFER
 */
static
void
_ContentsEnd(STRINGBUFFER *sbout)
{
  sbadd(sbout, 
        "<p><center><font size=2>Created using winhelpcgi &copy; 1996-2004 <a href=\"http://www.herdsoft.com\">Herd Software Development</a></font></center>\n"
        "\n</body>\n</html>\n");
}			 



/**
 *
 * Enumerate the Contents of the Help-File
 *
 * @param sbout                Output STRINGBUFFER
 * @param HlpFileName          Filename of the .hlp file to be enumerated
 * @param HlpTitleScan         Optional Keywords to be searched in all Titles
 * @param HlpKeyWord           Keyword to be searched in keyword list
 * @param HlpPage              Page number of the list to be shown
 * @param mimetype             On return contains the content-type
 * @param mimetypesize         Size of mimetype
 */
static
void
_EnumerateContents(STRINGBUFFER *sbout, LPCSTR HlpFileName, LPCSTR HlpTitleScan,
                   LPCSTR HlpKeyWord, LPCSTR HlpPage, char *mimetype, int mimetypesize)
{
  struct vchm *pchm;
  LPHELPFILETITLE lpHelpTitle;

  if (NULL!=(pchm=VCHMOpen(HlpFileName)) &&
      NULL!=(lpHelpTitle = HlpGetHelpTitle(VCHMGetIfsHandle(pchm)))
      )
    {
      char mimetype[100];
      STRINGBUFFER sbhhc;
      memset(&sbhhc, 0, sizeof(sbhhc));

      _ContentsHeading(sbout, HlpFileName, VCHMGetIfsHandle(pchm), lpHelpTitle, HlpTitleScan, HlpKeyWord);
      _ContentsStart(sbout, HlpFileName, VCHMGetIfsHandle(pchm), lpHelpTitle, HlpTitleScan, HlpKeyWord);

      if (VCHMGetFile(pchm, &sbhhc, "winhelpcgi.hhc", 0, mimetype, sizeof(mimetype)))
        {
          const char *src;
          char topictitle[1000];
          char local[150];
          local[0]=
            topictitle[0]=0;

          for (src=sbget(&sbhhc); *src; )
            {
              // Ignore whitespace
              while (isspace(*src)) src++;

              if (*src=='<')
                {
                  char par_name[100];
                  char par_value[1000];
                  char tagname[50];
                  char *dest;
                  src++;

                  par_name[0]=
                    par_value[0]=0;
                  
                  while (isspace(*src)) src++;

                  for (dest=tagname; dest-tagname < sizeof(tagname)-1 && *src && !isspace(*src) && *src!='>'; )
                    *dest++ = tolower(*src++);
                  *dest=0;

                  while (isspace(*src)) src++;

                  while (*src && *src!='>')
                    {
                      char paramname[50];
                      char paramvalue[1000];
                      for (dest=paramname; dest-paramname < sizeof(paramname)-1 && *src && !isspace(*src) && *src!='=' && *src!='>'; )
                        *dest++ = tolower(*src++);
                      *dest=0;
                      
                      while (isspace(*src)) src++;

                      if (*src=='=')
                        {
                          src++;
                          while (isspace(*src)) src++;
                          dest=paramvalue;
                          if (*src=='\"')
                            {
                              src++;
                              while (dest-paramvalue < sizeof(paramvalue)-1 && *src && *src!='\"')
                                *dest++ = *src++;
                              if (*src=='\"')
                                src++;
                            }
                          else if (*src=='\'')
                            {
                              src++;
                              while (dest-paramvalue < sizeof(paramvalue)-1 && *src && *src!='\'')
                                *dest++ = *src++;
                              if (*src=='\'')
                                src++;
                            }
                          else
                            {
                              while (dest-paramvalue < sizeof(paramvalue)-1 && *src && !isspace(*src))
                                *dest++ = *src++;
                            }
                          *dest=0;
                          
                          while (isspace(*src)) src++;
                        }
                      else
                        {
                          paramvalue[0]=0;
                        }

                      if (!strcmp(paramname, "value"))
                        {
                          strncpy(par_value, paramvalue, sizeof(par_value));
                        }
                      else if (!strcmp(paramname, "name"))
                        {
                          strncpy(par_name, paramvalue, sizeof(par_name));
                        }
                    }

                  if (!strcmp(tagname, "ul"))
                    {
                      sbprintf(sbout, "<ul id=\"winhelpcgi\">\n");
                    }
                  if (!strcmp(tagname, "/ul"))
                    {
                      sbprintf(sbout, "</ul>\n");
                    }
                  else if (!strcmp(tagname, "param"))
                    {
                      if (!strcasecmp(par_name, "name"))
                        {
                          strncpy(topictitle, par_value, sizeof(topictitle));
                        }
                      else if (!strcasecmp(par_name, "local"))
                        {
                          strncpy(local, par_value, sizeof(local));
                        }
                    }
                  else if (!strcasecmp(tagname, "/object"))
                    {
                      /* End of the object tag. Write collected informations. */

                      if (HlpTitleScan==NULL ||
                          HlpTitleScan[0]==0 ||
                          strstr(topictitle, HlpTitleScan))
                        {
                          if (local[0])
                            {
                              sbprintf(sbout, "<li class=\"booktopic\"><a href=\"%s\">",
                                     UrlEscape(local));
                            }
                          else
                            {
                              sbadd(sbout, "<li class=\"bookopen\">");
                            }
                          
                          sbhtmlputs(sbout, topictitle[0] ? topictitle : _("<No Title - maybe popup...>"));
                          
                          if (local[0])
                            {
                              sbadd(sbout, "</a>");
                            }
                          sbadd(sbout, "</li>\n");
                        }
                      
                      local[0]=
                        topictitle[0]=0;
                    }
                  
                  if (*src=='>')
                    {
                      src++;
                    }
                }
              else
                {
                  sbputchar(sbout, *src++);
                }
            }

          sbclear(&sbhhc);
        }

      HlpFreeHelpTitle(lpHelpTitle);
      VCHMClose(pchm);

      _ContentsEnd(sbout);
    }
  else
    {
      sbprintf(sbout, "<html>\n<body>\nError processing this file.\n</body>\n</html>\n");
    }

  strncpy(mimetype, "text/html", mimetypesize);
}

/**
 * Enumerate the Contents of the Help-File based on a search keyword
 *
 * @param sbout                Output STRINGBUFFER
 * @param HlpFileName          Filename of the .hlp file to be enumerated
 * @param HlpTitleScan         Optional Keywords to be searched in all Titles
 * @param HlpKeyWord           Keyword to be searched in keyword list
 * @param HlpPage              Page number of the list to be shown
 * @param mimetype             On return contains the content-type
 * @param mimetypesize         Size of mimetype
 */
void _EnumerateKeyword(STRINGBUFFER *sbout, LPCSTR HlpFileName, LPCSTR HlpTitleScan,
                       LPCSTR HlpKeyWord, LPCSTR HlpPage, char *mimetype, int mimetypesize)
{
  HIFS            ifs = IFSOpen(HlpFileName);
  TITLEDATA       tit;
  LPHELPFILETITLE lpHelpTitle;

  if (ifs)
	 {
		/* ----------- Get Help file Title -------------------*/
		lpHelpTitle = HlpGetHelpTitle(ifs);
		if (lpHelpTitle)
		  {
			 memset(&tit, 0, sizeof(tit));
			 
			 tit.HlpTitleScan=HlpTitleScan;
			 tit.HlpFileName=HlpFileName;
          tit.Page=HlpPage!=NULL ? atoi(HlpPage) : 0;
          tit.PageCount=(HlpGetTopicCount(ifs) + TITLES_PER_PAGE-1) / TITLES_PER_PAGE;
          tit.htopic = HlpTopicOpen(ifs, /*HTO_FONTS |*/ HTO_CONTEXT/* | HTO_KEYWORDS*/);
          tit.script_name = getenv("SCRIPT_NAME");
          tit.sbout = sbout;
			 
          // Render the first part
          _ContentsHeading(sbout, HlpFileName, ifs, lpHelpTitle, HlpTitleScan, HlpKeyWord);


          snprintf(tit.BaseUrl, sizeof(tit.BaseUrl),
                   HlpKeyWord != NULL
                   ? "%s/%s/?KEYWORD=%s"
                   : "%s/%s/",
                   tit.script_name,
                   HlpFileName, HlpKeyWord);

          sbprintf(sbout, "<link rel=\"start\" href=\"%s\" title=\"Start\">\n", tit.BaseUrl);
          if (tit.Page>0)
            {
              sbprintf(sbout, "<link rel=\"prev\" href=\"%s&PAGE=%d\" title=\"%s\">\n", tit.BaseUrl, tit.Page-1, _("Previous Page"));
            }
          if (tit.Page+1<tit.PageCount)
            {
              sbprintf(sbout, "<link rel=\"next\" href=\"%s&PAGE=%d\" title=\"%s\">\n", tit.BaseUrl, tit.Page+1, _("Next Page"));
            }

          // Render the first part of the text
          _ContentsStart(sbout, HlpFileName, ifs, lpHelpTitle, HlpTitleScan, HlpKeyWord);

			 /* -------- Enumerate the Contents of the Help File-*/
          DisplayKeyWordListing(sbout, ifs, &tit, HlpKeyWord);
			 
			 HlpFreeHelpTitle(lpHelpTitle);
          HlpTopicClose(tit.htopic);
          _ContentsEnd(sbout);
		  }
		else sbprintf(sbout, ERROR_FONT "Sorry, can't retrieve Title Data for %s" NORMAL_FONT, HlpFileName);
		
		IFSClose(ifs);
		
	 } else sbprintf(sbout, ERROR_FONT "Sorry, can't open File %s" NORMAL_FONT, HlpFileName);

  strncpy(mimetype, "text/html", mimetypesize);
}


/* _HtmlListHelpDirectoryContents
 *
 * List a Directory containing help files including all the subdirectories
 *
 * @param sbout           Output STRINGBUFFER
 */
static
void
_HtmlListHelpDirectoryContents(STRINGBUFFER *sbout, LPCSTR lpcszDirName)
{
  
  DIR           * dir=opendir(lpcszDirName);
  struct dirent * ent;
  HIFS           ifs;
  LPHELPFILETITLE tit;
  char            szFullPathName[MAX_PATH];
  LPCSTR          point;
  const char *    script_name = getenv("SCRIPT_NAME");
  
  if (dir)
	 {
		while (NULL!=(ent=readdir(dir)))
        {
			 if (strcmp(lpcszDirName, "."))
				{
				  sprintf(szFullPathName, 
							 "%s/%s",
							 lpcszDirName,
							 ent->d_name);
				}
			 else
				{
				  strcpy(szFullPathName, ent->d_name);
				}
			 
			 if (NULL!=(point=strrchr(ent->d_name, '.')) &&
				  (0==  lstrcmpi(point+1, "hlp")||0==lstrcmpi(point+1, "mvb")) &&
				  NULL!=(ifs = IFSOpen(szFullPathName))
				  )
            {
				  if (NULL!=(tit=HlpGetHelpTitle(ifs)))
                {
#ifdef FASTER
						sbprintf(sbout, "<tr><td><a href=\"%s/%s/topicoffset/0x%08x.html\">", script_name, szFullPathName, tit->ContentsTopicOffset);
#else
                  HTOPIC topic;
                  topic      = HlpTopicOpen(ifs, /*HTO_FONTS |*/ HTO_CONTEXT/* | HTO_KEYWORDS*/);
                  if (topic)
                    {
                      const char *ContentsContextID = HlpTopicFindContextIdForTopicOffset(topic, tit->ContentsTopicOffset);
                      if (ContentsContextID)
                        {
                          sbprintf(sbout, "<tr><td><a href=\"%s/%s/id/%s.html\">", script_name, szFullPathName, ContentsContextID);
                        }
                      else
                        {
                          sbprintf(sbout, "<tr><td><a href=\"%s/%s/topicoffset/0x%08x.html\">", script_name, szFullPathName, tit->ContentsTopicOffset);
                        }

                      HlpTopicClose(topic);
                    }
#endif
						sbhtmlputs(sbout, ent->d_name);
						
						sbadd(sbout, "</a></td><td>");
						sbhtmlputs(sbout, tit->HelpTitle ? tit->HelpTitle : "-");
						
						sbadd(sbout, "</td><td>");
						sbhtmlputs(sbout, tit->Copyright ? tit->Copyright : "-");
						
						sbadd(sbout, "</td><td>");
						sbhtmlputs(sbout, tit->Citation ? tit->Citation : "-");

                  sbprintf(sbout, "<td align=\"right\"<a href=\"%s/%s/\" title=\"%s\">%d</a></td>", 
                         script_name,
                         szFullPathName, 
                         _("Contents Listing"),
                         HlpGetTopicCount(ifs));

                  sbprintf(sbout, "<td align=\"right\"<a href=\"%s/%s/?KEYWORD=\" title=\"%s\">%d</a></td>", 
                         script_name,
                         szFullPathName, 
                         _("Keyword Search"),
                         HlpGetKeywordCount(ifs));

                  {
                    char buffer[500];
						  const char *format;
#ifdef HAVE_LANGINFO_H
						  format = nl_langinfo(D_FMT);
#else
						  format = "%d.%m.%Y";
#endif
                    strftime(buffer, sizeof(buffer), format, localtime(&tit->GenDate));
                    sbprintf(sbout, "<td>%s</td>", buffer);
                  }
						
						sbadd(sbout, "</tr>\n");
						HlpFreeHelpTitle(tit);
                }
				  
				  IFSClose(ifs);
            }
			 else
				if (ent->d_name[0]!='.')
				  {
					 _HtmlListHelpDirectoryContents(sbout, szFullPathName);
				  }
        }
		
		closedir(dir);
	 } // if (dir)
  
} // _HtmlListHelpDirectoryContents


/* _HtmlUsage
 *
 * Explain the Usage of this Program as an HTML-Document
 *
 * @param sbout                Output STRINGBUFFER
 * @param mimetype             On return contains the content-type
 * @param mimetypesize         Size of mimetype
 */
static
void
_HtmlUsage(STRINGBUFFER *sbout, char *mimetype, int mimetypesize)
{
  const char *script_name = getenv("SCRIPT_NAME");
  
  sbprintf(sbout, "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
           "<html>\n"
           "<head>\n"
           "<title>%s</title>\n",
           _("Accessing a Windows Help File"));

  sbadd(sbout, "</head>");

  sbprintf(sbout,
           "<h1>%s</h1>"
           "<p>%s</p>\n"
           "<hr>"
           "<h2>%s</h2>"
           "<form method=GET action=\"%s\">\n"
           "<p>%s:<INPUT NAME=\"KEYWORD\" size=\"50\" value=\"\"></p>\n"
           "<p><INPUT TYPE=SUBMIT value=\"%s\"></p>"
           "</form>"
           "<hr>"
           
           "<h2>%s</h2>",
           _("Accessing a Windows Help File"),
           _("This CGI-Program allows reading of a Windows Help (WinHelp) file or Media "
             "Viewer 2.0 file from your favorite web Browser."),
           _("Search for a keyword in all available files"),
           script_name,
           _("KeyWord"),
           _("Start Search"),
           _("List of available help files")
           );
  
  sbprintf(sbout, "<table border=1><tr><th>%s</th><th>%s</th><th>%s</th><th>Citation</th><th>%s</th><th>%s</th><th>%s</th></tr>",
           _("Filename"),
           _("File Title"),
           _("Copyright"),
           _("Topics"),
           _("Keywords"),
           _("Date"));
  
  
  _HtmlListHelpDirectoryContents(sbout, ".");
#ifdef __netware__
  _HtmlListHelpDirectoryContents(sbout, "sys:login");
  _HtmlListHelpDirectoryContents(sbout, "sys:public");
#endif
  
  sbadd(sbout, "</table>"
        "<p>Copyright &copy; 1996-2004 Herd Software Development <a href=\"http://www.herdsoft.com\">http://www.herdsoft.com</a>\n"
        );
  
  sbadd(sbout, "\n</body>\n</html>\n");


  /* -------- MIME Content-Type ----------------------*/
  strncpy(mimetype, "text/html", mimetypesize);
} // _HtmlUsage


/**
 * Check if a given filename can be considered secure, that means it does
 * not point into a directory above the current directory.
 *
 * For OS with unnormal filesystem semantics this must be modified
 */
static
BOOL
_IsSecureFileName(const char * HlpFileName)
{
#ifdef __netware__
  return !strchr(HlpFileName, ':') && !strchr(HlpFileName, '/') && strchr(HlpFileName, '\\');
#else
  /* What a clean concept Posix is compared to Windows...! */
  return HlpFileName[0] != '/' && strstr(HlpFileName, "../")==NULL;
#endif
}



/* ============= Main Program Control =========================================*/


/* WinHelpCGI Main Program
 *
 */
void WinHelpCGI(LPCSTR QueryString)
{
  const char* HlpFileName = NULL;
  const char* HlpKeyWord;
  const char* HlpTitleScan;
  const char* HlpPage;
  const char* virtname = NULL;

  const char* PathInfo;

  char mimetype[50];
  time_t gendate = time(NULL);

  STRINGBUFFER sbout;
  memset(&sbout, 0, sizeof(sbout));
  strncpy(mimetype, "text/plain", sizeof(mimetype));
  
  HlpTitleScan   = GetQueryVariable(QueryString, "TITLESCAN=");
  HlpKeyWord     = GetQueryVariable(QueryString, "KEYWORD=");
  HlpPage        = GetQueryVariable(QueryString, "PAGE=");

  /* Search the point in the PATH_INFO, where the physical filename ends
   * and the virtual name begins.
   */
  PathInfo = getenv("PATH_INFO");
  if (PathInfo != NULL)
    {
      char filename[500];
      char *dest;
      BOOL foundfile=FALSE;

      // Ignore leading '/'
      if (*PathInfo=='/')
        PathInfo++;

      for (dest=filename; *PathInfo && (dest-filename)<sizeof(filename)-1 && !foundfile; )
        {
          if (*PathInfo=='/')
            {
              struct stat sbf;
              *dest=0;
              if (stat(filename, &sbf)==0 && S_ISREG(sbf.st_mode))
                {
                  /* We've got a name that is a regular file, but is followed by a '/', 
                   * that must be the name of the help file.
                   */
                  foundfile = TRUE;

                  PathInfo++;
                }
              else
                {
                  *dest++ = *PathInfo++;
                }
            }
          else
            {
              *dest++ = *PathInfo++;
            }
        }

      if (foundfile)
        {
          HlpFileName = strdup(filename);
          if (*PathInfo=='/')
            {
              PathInfo++;
            }
          
          virtname = strdup(PathInfo);
        }
    }

  if (HlpFileName)
	 {
      if (!_IsSecureFileName(HlpFileName))
        {
          sbprintf(&sbout,
                   _("The filename specified is considered insecure.\n"
                     "Please copy or symlink files into the directory containing winhelpcgi.cgi at \"%s\"\n"),
                   getenv("SCRIPT_NAME")
                   );
        }
		else if (virtname && virtname[0])
        {
          struct vchm *pchm;

          if (NULL!=(pchm=VCHMOpen(HlpFileName)))
            {
              _VirtualAnswer(&sbout, pchm, virtname, mimetype, sizeof(mimetype));
              gendate = VCHMGetGenDate(pchm);
  
              VCHMClose(pchm);
            } 
          else
            {
              sbprintf(&sbout, _("%s Sorry, can't open File %s reason: %s %s"), ERROR_FONT, HlpFileName, strerror(errno), NORMAL_FONT);
            }
        }
      else
        {
          if (HlpKeyWord && HlpKeyWord[0])
            {
              _EnumerateKeyword(&sbout, HlpFileName, HlpTitleScan, HlpKeyWord, HlpPage, mimetype, sizeof(mimetype));
            }
          else
            {
              _EnumerateContents(&sbout, HlpFileName, HlpTitleScan, HlpKeyWord, HlpPage, mimetype, sizeof(mimetype));
            }
        }
    }
  else
    {
      /* No HelpFile Name specified yet. */
      if (HlpKeyWord && HlpKeyWord[0])
		  SearchKeyWordInAllHelpFiles(&sbout, HlpKeyWord, mimetype, sizeof(mimetype));
      else
		  _HtmlUsage(&sbout, mimetype, sizeof(mimetype));
    }

  /* Post-process HTML-Files. */
  if (!strcmp(mimetype, "text/html"))
    {
      if (virtname==NULL ||
          (memcmp(virtname, "print", 5) &&
           memcmp(virtname, "popup", 5)))
        {
          PostProcess(&sbout, HlpFileName);
        }
      GzipEncode(&sbout);
    }


  if (strcmp(mimetype, "hack/redirect"))
    {
      printf("Content-Type: %s\n"
             "Content-Length: %d\n"
             "Last-Modified: %s\n"
             "\n",
             mimetype,
             sbout.pos,
             _TimeString( gendate )
             );
    }

  sbflush(&sbout, stdout);

} // WinHelpCGI




/**
 * Set the gettext domain based on the browser language informations.
 */
static
void
_ChooseTextDomain(void)
{
  char language[10];
  if (ChooseLanguage(language, sizeof(language)) )
    {
      if (!strcasecmp(language, "de"))
        {
          setlocale (LC_TIME, "de_DE");
          setlocale (LC_CTYPE, "de_DE");
#ifdef ENABLE_NLS
          setlocale (LC_MESSAGES, "de_DE");
          bindtextdomain (PACKAGE, "share/locale");
          textdomain (PACKAGE);
#endif
        }
      else if (!strcasecmp(language, "nl"))
        {
          setlocale (LC_TIME, "nl_NL");
          setlocale (LC_CTYPE, "nl_NL");
#ifdef ENABLE_NLS
          setlocale (LC_MESSAGES, "nl_NL");
          bindtextdomain (PACKAGE, "share/locale");
          textdomain (PACKAGE);
#endif
        }
      else if (!strcasecmp(language, "fr"))
        {
          setlocale (LC_TIME, "fr_FR");
          setlocale (LC_CTYPE, "fr_FR");
#ifdef ENABLE_NLS
          setlocale (LC_MESSAGES, "fr_FR");
          bindtextdomain (PACKAGE, "share/locale");
          textdomain (PACKAGE);
#endif
        }
    }
}



/**
 * Encode given string in a way so it will always fit into an URL string.
 *
 * @param plain    Plaintext string
 */
const char *UrlEscape(const char *plain)
{
  static char buffer[5][200];
  static int run;
  char *ptr = buffer[(run++%5)];
  char *dest;
  for (dest=ptr; *plain && (dest-ptr)<sizeof(buffer[0])-4; plain++)
    {
      if ((*plain>='a' && *plain<='z') ||
          (*plain>='A' && *plain<='Z') ||
          (*plain>='0' && *plain<='9') ||
          strchr("._/", *plain)!=NULL
          )
        {
          *dest++=*plain;
        }
      else if (*plain==' ')
        {
          *dest++=' ';
        }
      else
        {
          sprintf(dest, "%%%02X", *plain);
          dest += 3;
        }
    }
  *dest=0;
  return ptr;
}



int
main(int argc, char **argv)
{ 
  const char *QueryString=getenv("QUERY_STRING");

  if (QueryString && NULL!=getenv("SCRIPT_NAME"))
	 { 
      _ChooseTextDomain();

		WinHelpCGI(*QueryString=='\"'?QueryString+1:QueryString);
	 }
  else if(argc>1 && !strcmp(argv[1], "--tar"))
    {
      Hlp2Tar(argc-2, argv+2);
    }
  else
	 {
		printf("winhelpcgi: " VERSION " CGI program used to display Windows Help (WinHelp .HLP) files\n"
             "            from HTML pages using a web browser.\n\n"
             "USAGE   :   This Program is intented to be used by a HTTP (WWW-) Server\n"
             "            for Internet Purposes.\n\n"
             "            it is based on the CGI (Common Gateway Interface) standard\n"
             "            for Internet Programming\n\n"
             "            You need to install the " CGI_NAME " program into a CGI-Enabled\n"
             "            directory of your (apache?) web server\n\n"
             "(C)Copyright 1996-2004 Herd Software Development\n"
             "Kettelerstr. 35 / 68642 Buerstadt / Germany / http://www.herdsoft.com\n");
	 }
  
  return 0;
} // main


/*
 * Local variables:
 *  tab-width: 3
 * End:
 */
